unit data3d; (* 3D 0.4 *)

{ Questa unita' riunisce elementi comuni ai seguenti programmi	}
{ 		FILE3D.PAS					}
{		GRAPH3D.PAS					}
{		3D.PAS						}
{ 12/94 Rev 0.1 : added *loop*                                  }
{ 12/94 Rev 0.2 : added *color*,crossv                          }
{ 10/95 Rev 0.3 : added *PolyCx*                                }
{ 11/95 Rev 0.4 : i record tClass e Modello sono stati fusi in tModello }
{ 12/97 Rev 0.45: added procedure disposetree                   }

interface
const n_max_vertici=512;
      n_max_superfici=128;
{      tethav=1.047;  angolo di visuale espresso in radianti }
      d_min=0.01;   { distanza minima da focalizzare (l=1/2) }
      epsilon=1e-7; { usato nei confronti con lo zero per    }
		    { problemi di overflow		     }
      epsilon2=epsilon*epsilon;
{ La costante <normal_factor> viene utilizzata per stabilire se
  due versori sono ortogonali. I versori n1 e n2 sono ritenuti
  ortogonali se abs(n1.n2)<normal_factor. Questa costante e' necessaria
  per ridurre gli effetti delle approssimazioni di calcolo e
  dei dati in ingresso }
      normal_factor=0.001;

  const right=8;left=4;
	down=2;up=1;
	retro=16;

const zMaxPolyCxVertex=16;
type pPolyCx=^tPolyCx;    {*PolyCx*}
     tPolyCx=record
                 vertex:array[1..zMaxPolyCxVertex] of word;  (* vertex *)
                 n_vertex:integer; (* number of vertex *)
                 next:pPolyCx;
               end;
type pclass=^class;
     class=record
	    vertice:word;
	    next:pclass;
	    cont:boolean;
	    end;
type tpoint=record
			x,y,z:real;
		  end;
     tproiezione=    record
			xp,yp:integer;
			rel:tpoint;
			status:byte;
		     end;
     colore=record
             r,g,b:byte;
            end;
     tmaterial=record
                 ambientColor,
                 diffuseColor,
                 specularColor,
                 emissiveColor:colore
               end;
     pMaterial=^tmaterial;
     pLight=^tLight;
     tLight=record
                    color:colore;
                    dir:tpoint;
                    next:pLight;
                  end;
     vettore_di_colori=array[1..n_max_superfici]of pMaterial;
     vettore_di_liste_convessi=array[1..n_max_superfici]of pPolyCx;
     vettore_di_vertici=array[0..n_max_vertici-1]of tpoint;
     vettore_proiezioni=array[0..n_max_vertici-1]of tproiezione;
     vettore_di_normali=array[1..n_max_superfici]
					of tpoint;
     vettore_sup_luminoso=array[1..n_max_superfici]of colore;
     vettore_sup_convessa=array[1..n_max_superfici]of boolean;
     idx_sup=array[1..n_max_superfici]of integer; {*loop*}

type vettore_di_superfici=array[1..n_max_superfici]of pclass;
     vettore_concavi=array[1..n_max_superfici]of pclass;
     plist_sup=^list_sup;
     list_sup=record
                sup:integer;
                next:plist_sup
              end;
     vettore_sup_coperte=array[1..n_max_superfici]of plist_sup;
     pmodello=^tmodello;
     pPointArray=^tPointArray;
     tPointArray=record
                   min,max:tpoint; (* Bounding Box *)
	           middle:tpoint;
	           vertici:vettore_di_vertici;
   	           n_vertici:word;
                   pmastermdl:pmodello;  (* used for master *)
                 end;
     tbase=   object
              extension:real;
              PointArrayp:pPointArray;
	      n_superfici:word;			{ usato nel modello di superfici }
              end;
     tmodello=object(tbase)
              list_PolyCxs: vettore_di_liste_convessi;(*PolyCx*) { usato nel modello di superfici }
	      normali:vettore_di_normali;       { usato nel modello di superfici }
	      superfici:vettore_di_superfici;   { usato nel modello di superfici }
	      convessa:vettore_sup_convessa;    { usato nel modello di superfici }
              colori:vettore_di_colori; (*color*){ usato nel modello di superfici }
(* NOTA: il primo elemento di <colori> e' usato anche per le spezzate ! *)
              loop:boolean; {*loop*}
              _last:integer; {*loop*}
              sup_coperte:^vettore_sup_coperte; {*loop*}
	     end;

     pline=^tline;
     tline=  object(tbase)
              classe:pclass;
              colori:vettore_di_colori; (*color*){ usato nel modello di superfici }
(* NOTA: il primo elemento di <colori> e' usato anche per le spezzate ! *)
	     end;
     tstate=record
	      posizione:tpoint;
	      cs,sn:real;
	      scala:real;
	    end;
     ttypenode=(lineset,faceset,structure);
     pnode=^tnode;
     tnode=record
             next:pnode;
             state:tstate;
             attivo:boolean;
             case typenode: ttypenode of
               lineset:(line:pline);
               faceset:(tipo:pmodello);
               structure:(child:pnode);
           end;

(* VARIABILI GLOBALI *)
var sky,ground:colore; {colori dello sfondo (cielo e terra)}

(* begin currentState *)
    currentLight:pLight;
    defaultMaterial:pMaterial;
(* end currentState *)

procedure duplicatetree(source:pnode;var dest:pnode);
procedure disposetree(source:pnode);
procedure cross(var ris:tpoint;a,b,c:tpoint);
procedure crossv(var ris:tpoint;var a,b:tpoint);
procedure versore(var q:tpoint);
procedure Increment_Position(pobj:pnode;dx,dy,dz:real);
procedure Set_Position(pobj:pnode;pos:tpoint);
procedure Get_Position(pobj:pnode;var pos:tpoint);
procedure Move_Position(pobj:pnode;delta:real);
procedure Set_Rotation2(pobj:pnode;ca,sa:real);
procedure Rotate(pobj:pnode;alfa:real);
procedure Increment_Scale(pobj:pnode;dscale:real);
procedure Set_Scale(pobj:pnode;newscale:real);
function dot3(a,b,c:tpoint):real;
function dot(a,b:tpoint):real;
function distance(var a,b:tpoint):real;
function magnitude(var a:tpoint):real;
procedure diff(var a,b,c:tpoint);
procedure sum(var a,b,c:tpoint);
procedure scalar(var a,b:tpoint;m:real);
function PointIsIncluded(var point:tpoint;p:pnode;phi:real):pnode;
function DetectCollision(pobj,p:pnode):pnode;

implementation

procedure duplicatetree(source:pnode;var dest:pnode);
{ copy node and related sub-tree }
var pchild:pnode;
begin
  new(dest);
  dest^:=source^;
  if source^.typenode = structure
  then begin
       duplicatetree(source^.child,dest^.child);
       source:=source^.child^.next;pchild:=dest^.child;
       while source <> nil
       do begin
            duplicatetree(source,pchild^.next);
            source:=source^.next;
            pchild:=pchild^.next;
          end;
       end;
end;

procedure disposetree(source:pnode);
{ free node and related sub-tree }
begin
  if source^.next <> nil
  then disposetree(source^.next);
  if source^.typenode = structure
  then disposetree(source^.child);
  dispose(source)
end;

procedure cross(var ris:tpoint;a,b,c:tpoint);
{ compute the cross product }
{ RIS := ( A - B ) x ( C - B ) }
begin
   a.x:=a.x-b.x;a.y:=a.y-b.y;a.z:=a.z-b.z;
   c.x:=c.x-b.x;c.y:=c.y-b.y;c.z:=c.z-b.z;
   ris.x:=(a.y*c.z-a.z*c.y);
   ris.y:=(-a.x*c.z+a.z*c.x);
   ris.z:=(a.x*c.y-a.y*c.x);
end; { of cross }

procedure crossv(var ris:tpoint;var a,b:tpoint);
{ compute the cross product }
{ RIS := A x B }
begin
   ris.x:=(a.y*b.z-a.z*b.y);
   ris.y:=(-a.x*b.z+a.z*b.x);
   ris.z:=(a.x*b.y-a.y*b.x);
end; { of cross }

procedure scalar(var a,b:tpoint;m:real);
{ compute the vector multiplication by a scalar }
{ A := mB }
begin
   a.x:=m*b.x;a.y:=m*b.y;a.z:=m*b.z;
end; { of scalar }

procedure sum(var a,b,c:tpoint);
{ compute the sum of vectors }
{ A := B + C }
begin
   a.x:=b.x+c.x;a.y:=b.y+c.y;a.z:=b.z+c.z;
end; { of sum }

procedure diff(var a,b,c:tpoint);
{ compute the vector difference }
{ A := B - C }
begin
   a.x:=b.x-c.x;a.y:=b.y-c.y;a.z:=b.z-c.z;
end; { of diff }

function dot(a,b:tpoint):real;
{ compute the dot product }
{ DOT := A . B }
begin
   dot:=a.x*b.x+a.y*b.y+a.z*b.z;
end; { of dot }


function dot3(a,b,c:tpoint):real;
{ compute the following dot product }
{ DOT3 := ( A - B ) . ( C - B ) }
begin
   a.x:=a.x-b.x;a.y:=a.y-b.y;a.z:=a.z-b.z;
   c.x:=c.x-b.x;c.y:=c.y-b.y;c.z:=c.z-b.z;
   dot3:=dot(a,c);
end; { of dot3 }

procedure versore(var q:tpoint);
{ Q := Q / MAG(Q) }
var mag:real;
begin
with q
do begin
   mag:=sqrt(sqr(x)+sqr(y)+sqr(z));
   x:=x/mag;
   y:=y/mag;
   z:=z/mag
   end
end; { of versore }

function magnitude(var a:tpoint):real;
begin
  magnitude:=sqrt(sqr(a.x)+sqr(a.y)+sqr(a.z));
end; { of magnitude }

function distance(var a,b:tpoint):real;
{ compute the distance between A and B }
begin
  distance:=sqrt(sqr(a.x-b.x)+sqr(a.y-b.y)+sqr(a.z-b.z));
end; { of distance }

{ Attenzione: una qualsiasi applicazione non deve agire sui contenuti	}
{		delle strutture dati di questo programma MA deve fare	}
{		uso delle seguenti procedure !!!			}

procedure Increment_Position(pobj:pnode;dx,dy,dz:real);
var child:pnode;
begin
  with pobj^.state.posizione
  do begin
       x:=x+dx;y:=y+dy;z:=z+dz;
       if pobj^.typenode = structure
       then begin
              child:=pobj^.child;
              while child <> nil
              do begin
                   Increment_Position(child,dx,dy,dz);
                   child:=child^.next;
                 end;
            end;
     end;
end;

procedure Set_Position(pobj:pnode;pos:tpoint);
begin
  with pobj^.state.posizione
  do Increment_Position(pobj,pos.x-x,pos.y-y,pos.z-z);
end; { of Set_Position }

procedure Get_Position(pobj:pnode;var pos:tpoint);
begin
with pobj^.state.posizione
do begin
   pos.x:=x;
   pos.y:=y;
   pos.z:=z;
   end
end; { of Get_Position }

procedure Move_Position(pobj:pnode;delta:real);
var child:pnode;
begin
with pobj^.state
do Increment_Position(pobj,delta*cs,delta*sn,0);
end; { of Move_Position }

procedure Rotate2(pobj:pnode;ca,sa:real;xc,yc:real);
var temp:real;
var child:pnode;
begin
with pobj^.state,posizione do
    begin
	temp:=cs*ca-sn*sa;
	sn:=sn*ca+cs*sa;
	cs:=temp;
        temp:=(x-xc)*ca-(y-yc)*sa+xc;
        y:=(y-yc)*ca+(x-xc)*sa+yc;
        x:=temp;
        if pobj^.typenode = structure
        then begin
               child:=pobj^.child;
               while child <> nil
               do begin
                    Rotate2(child,ca,sa,xc,yc);
                    child:=child^.next;
                  end;
             end;
    end
end; { of Rotate }

procedure Rotate(pobj:pnode;alfa:real);
var temp:real;
var child:pnode;
begin
  with pobj^.state.posizione
  do Rotate2(pobj,cos(alfa),sin(alfa),x,y);
end; { of Rotate }


procedure Set_Rotation2(pobj:pnode;ca,sa:real);
var dcs,dsn:real;
begin
with pobj^.state,posizione do
     begin
	dcs:=cs*ca+sn*sa;
	dsn:=-sn*ca+cs*sa;
        Rotate2(pobj,dcs,dsn,x,y);
     end
end; { of Set_Rotation }


procedure Increment_Scale(pobj:pnode;dscale:real);
var child:pnode;
begin
  with pobj^.state
  do begin
       scala:=dscale*scala;
       if pobj^.typenode = structure
       then begin
              child:=pobj^.child;
              while child <> nil
              do begin
                   Increment_Scale(child,dscale);
                   child:=child^.next;
                 end;
            end;
     end;
end;

procedure Set_Scale(pobj:pnode;newscale:real);
var child:pnode;
begin
   Increment_Scale(pobj,newscale/pobj^.state.scala);
end; { of Set_Scale }

function DetectCollisionFaceSet(pobj,p:pnode):boolean;
(* questa procedura controlla l'eventuale collisione
   del faceset <p> l'oggetto <pobj> *)
var cl,sl:real;
    dx,dy,xr,yr,zr:real;
    point:tpoint;estensione:real;
begin
point:=p^.state.posizione;
with p^,state
do estensione:=scala*tipo^.extension;
while pobj<>nil
do with pobj^,point,state
   do begin
      if attivo
      then case typenode of
           structure:begin
                       if DetectCollisionFaceSet(pobj^.child,p)
                       then begin
                              DetectCollisionFaceSet:=true;
                              exit;
                            end;
                     end;
           faceset:
             begin
		dx:=posizione.x-x;
		dy:=posizione.y-y;
		xr:=dx;
		yr:=dy;
		zr:=posizione.z-z;
	        cl:=cs*scala;
		sl:=-sn*scala;
                with tipo^.PointArrayp^
                do begin
                   xr:=xr+middle.x*cl+middle.y*sl;
                   yr:=yr+middle.y*cl-middle.x*sl;
                   zr:=zr+middle.z*scala;
                   end;
		if sqr(xr)+sqr(yr)+sqr(zr) < sqr(scala*tipo^.extension+estensione)
                then begin
                       DetectCollisionFaceSet:=true;
                       exit;
                     end;
            end;
        end; (* of case *)
      pobj:=pobj^.next;
      end;
DetectCollisionFaceSet:=false;
end; { of DetectCollisionFaceSet }

function DetectCollision(pobj,p:pnode):pnode;
(* questa procedura cerca un oggetto appartenente a <p>
   che collide con l'oggetto <pobj> *)
var q:pnode;

begin
while p <> nil
do begin
   if p <> pobj
   then    begin
           if p^.attivo
           then case p^.typenode of
                structure:begin
                            q:=DetectCollision(pobj,p^.child);
                            if q <> nil
                            then begin
                                   DetectCollision:=q;
                                   exit;
                                 end;
                          end;
                faceset:    if DetectCollisionFaceSet(pobj,p)
                            then begin
                                   DetectCollision:=p;
                                   exit;
                                 end;
             end; (* of case *)
           end;
   p:=p^.next;
   end;
DetectCollision:=p;
end; { of DetectCollision }

function PointIsIncluded(var point:tpoint;p:pnode;phi:real):pnode;
(* questa procedura cerca un oggetto che include il punto <point> *)
var cf,sf,cl,sl:real;
    dx,dy,xr,yr,zr:real;
    q:pnode;
begin
cf:=cos(phi);
sf:=sin(phi);
while p<>nil
do with p^,point,state
   do begin
      if attivo
      then case typenode of
           structure:begin
                       q:=PointIsIncluded(point,p^.child,phi);
                       if q <> nil
                       then begin
                              PointIsIncluded:=q;
                              exit;
                            end;
                     end;
           faceset:
             begin
		dx:=posizione.x-x;
		dy:=posizione.y-y;
		xr:=dx*cf+dy*sf;
		yr:=dy*cf-dx*sf;
		zr:=posizione.z-z;
	        cl:=(cs*cf+sn*sf)*scala;
		sl:=(sf*cs-sn*cf)*scala;
                with tipo^.PointArrayp^
                do begin
                   xr:=xr+middle.x*cl+middle.y*sl;
                   yr:=yr+middle.y*cl-middle.x*sl;
                   zr:=zr+middle.z*scala;
                   end;
		if sqr(xr)+sqr(yr)+sqr(zr) < sqr(scala*tipo^.extension)
                then begin
                       PointIsIncluded:=p;
                       exit;
                     end;
            end;
        end; (* of case *)
      p:=p^.next;
      end;
PointIsIncluded:=p;
end; { of PointIsIncluded }


begin
(* initialize global variables *)
  sky.r:=0;
  sky.g:=0;
  sky.b:=0;
  ground:=sky;
  currentLight:=nil;
  new(defaultMaterial);
  with defaultMaterial^
  do begin
       with diffuseColor
       do begin
          r:=200;g:=200;b:=200
          end;
       with ambientColor
       do begin
          r:=50;g:=50;b:=50
          end;
       with specularColor
       do begin
          r:=0;g:=0;b:=0
          end;
       with emissiveColor
       do begin
          r:=0;g:=0;b:=0
          end;
     end;
end.

{nota: a programma terminato far dipendere questa unit dalla unit GRAPH3D }
{      e spostare la variabile graphic in GRAPH3D.PAS			  }
