unit graph3D; (* 3D 0.45 *)
{ Unita' utilizzabile solo su EGA,VGA e SVGA                    }
{$B-}
{$define hline}
{$define debug}
{define stat}
{ simboli riconosciuti dal compilatore: 			}
{	hline abilita la linea di orizzone			}
{	debug abilita la modalita' di debug			}
{	stat abilita la collezione di dati in stat[]            }

{ 23/8/1990 Clipping esteso alle superfici convesse		}
{ 22/9/1990 Introdotto il SORT nella routine SHOW()		}
{ 28/10/1990 Clipping esteso alle superfici concave		}
{ 1/12/1991 Rimosso errore                                      }
{ 12/94 Rev 0.1 : added *loop*, visita_spezzata                 }
{                 <conv> e' scritta in assembler                }
{ 12/94 Rev 0.2 : added *color*                                 }
{ 10/95 Rev 0.3 : added *256*                                   }
{  1/96 Rev 0.4 : modalita' 256 colori con accesso diretto      }
{ 12/97 Rev 0.45: adattamento a Windows. Vedi simbolo WINDOWS   }
(* NOTA: le routines <sort2> e <sup_copre2> sono poco efficienti
         modificare gli algoritmi usati per renderle efficienti
         si potrebbe agire in particolare su <idx2sup> e
         sulla gestione del clipping                             *)
(* NOTA: la routine <sup_visibile> non funziona quando tutti i
         vertici del poligono sono esterni all'immagine ma il
         poligono e' visibile                                    *)
(* Nota: BW marca le linee necessarie ad una visualizzazione a 64 grigi *)

interface
{$IFDEF WINDOWS}
  uses WObjects, Wintypes, Winprocs, WVGA256, Data3D;
  type TGraph3DWindow =
          object(TBufferWindow)
            procedure show(p:pnode;xo,yo,zo:real;phi:real;viewangle:real);
            constructor Init(AParent: PWindowsObject; Title: PChar);
          private
            procedure error(s:pChar);
            procedure SetFillColor(color:colore);
            procedure visita_spezzata(linep:pline;var lista:vettore_proiezioni);
            procedure FillSurface(tipo:pmodello;var lista:vettore_proiezioni;
				sup:integer; var sup_lum:vettore_sup_luminoso);
          end;

{$ELSE}
  uses Vga256,Graph,Data3D;

  type tmode = (color16,color256,gray64);
  var graphic:boolean;  {(vedi nota)}

  procedure opengraph(cmode:tmode); (*256*)
  procedure closegraph;
  procedure show(p:pnode;xo,yo,zo:real;phi:real;viewangle:real);
{  procedure TestPattern(cmode:tmode);}
  const
    FULL = 12;
    attFull = 255 div FULL;
    Gray : array[0..FULL]of FillPatternType =
    ((0,0,0,0,0,0,0,0),
     (0,$20,0,0,0,2,0,0),
     ($20,0,2,0,$80,0,8,0),
     ($20,2,$80,8,$20,2,$80,8),
     ($44,$11,$44,$11,$44,$11,$44,$11),
     ($AA,$44,$AA,$11,$AA,$44,$AA,$11),
     ($55,$AA,$55,$AA,$55,$AA,$55,$AA),
     ($55,$BB,$55,$EE,$55,$BB,$55,$EE),
     ($BB,$EE,$BB,$EE,$BB,$EE,$BB,$EE),
     ($DF,$FD,$7F,$F7,$DF,$FD,$7F,$F7),
     ($DF,$FF,$FD,$FF,$7F,$FF,$F7,$FF),
     ($FF,$DF,$FF,$FF,$FF,$FD,$FF,$FF),
     ($FF,$FF,$FF,$FF,$FF,$FF,$FF,$FF));
{$ENDIF}

implementation
function max(a,b,c:integer):integer;
begin
if a > b
then if a > c
     then max:=a
     else max:=c
else if b > c
     then max:=b
     else max:=c
end;

function IsBright(color:colore):boolean;
(* questa funzione verifica che il colore non e' troppo scuro *)
begin
  with color
  do IsBright:=max(r,g,b) >= 4;
end;

{$IFDEF WINDOWS}
const hmaxx=320 div 2-1;
      hmaxy=200 div 2-1;

function convert_color(r,g,b:byte):integer;
begin
  convert_color:=b div 64+(g div 8)and $FC+r and $E0;
end;

procedure TGraph3DWindow.SetFillColor(color:colore);
begin
  with color
  do TBufferWindow.SetFillColor(convert_color(r,g,b));
end;

procedure TGraph3DWindow.error(s:pChar);
begin
  MessageBox(HWindow, s, nil, mb_Ok)
end;

constructor TGraph3DWindow.Init(AParent: PWindowsObject; Title: PChar);
var r,g,b: integer;
    Palette: PaletteType;
begin
  TBufferWindow.Init(AParent, Title);
  for r:=0 to 7
  do for g:=0 to 7
     do for b:=0 to 3
        do begin
             Palette[r*32+g*4+b].r:=r*8;
             Palette[r*32+g*4+b].g:=g*8;
             Palette[r*32+g*4+b].b:=b*16;
           end;
  palette256(Palette);
end;

{$ELSE}
(* definizione delle procedure grafiche di base *)
var ClearDevice:procedure;
    Bar:procedure (x1, y1, x2, y2, color: Integer);
    Line:procedure (x1, y1, x2, y2, color: Integer);
    FillPoly:procedure (NumPoints: Word; var PolyPoints);


var multipage:boolean;
    pageused:integer;
    colormode:tmode;
    hmaxx,hmaxy:integer;
    buffer256:pBufferType;

{$ifdef stat}
const max_stat=5;
var stat:array[0..max_stat]of longint;
    istat:integer;
{$endif}

type (*256*)
  RGBRec = record
    RedVal, GreenVal, BlueVal : Integer;
     { Intensity values (values from 0 to 63) }
  end;
const (*256*)
  { Table of suggested colors for VGA 16 color modes }
  Colors : array[0..MaxColors] of RGBRec = (
   ( RedVal:8;GreenVal:8;BlueVal:8),
   ( RedVal:24;GreenVal:24;BlueVal:24),
   ( RedVal:40;GreenVal:8;BlueVal:8),
   ( RedVal:56;GreenVal:24;BlueVal:24),
   ( RedVal:8;GreenVal:40;BlueVal:8),
   ( RedVal:24;GreenVal:56;BlueVal:24),
   ( RedVal:40;GreenVal:40;BlueVal:8),
   ( RedVal:56;GreenVal:56;BlueVal:24),
   ( RedVal:8;GreenVal:8;BlueVal:40),
   ( RedVal:24;GreenVal:24;BlueVal:56),
   ( RedVal:40;GreenVal:8;BlueVal:40),
   ( RedVal:56;GreenVal:24;BlueVal:56),
   ( RedVal:8;GreenVal:40;BlueVal:40),
   ( RedVal:24;GreenVal:56;BlueVal:56),
   ( RedVal:40;GreenVal:40;BlueVal:40),
   ( RedVal:56;GreenVal:56;BlueVal:56)
  );

function convert_color(r,g,b:byte):integer;
var col:integer;
begin
  col:=0;
  case colormode of
  color256:begin
           col:=b div 64+(g div 8)and $FC+r and $E0;
           end;
  gray64: begin
           col:=b div 36+g div 7+r div 13;
        end;
  color16: begin
       if r >= 128
       then begin
            inc(col,2);
            r:=r-128;
            end;
       if g >= 128
       then begin
            inc(col,4);
            g:=g-128;
            end;
       if b >= 128
       then begin
            inc(col,8);
            b:=b-128;
            end;
       if (r+b+g) >= 192
       then begin
            inc(col);
            end;
       end;
  end;
convert_color:=col;
end;

procedure SetFillColor(color:colore);
var ma,amp:integer;

begin
  if colormode=color16
  then with color
       do begin
          ma:=max(r,g,b);
          if ma < 4
          then begin
               graph.SetFillStyle(SolidFill,black);
               graph.SetColor(black);
               graph.SetLineStyle(SolidLn,0,NormWidth);
               end
          else begin
               amp:=255 div ma;
               r:=r*amp;
               g:=g*amp;
               b:=b*amp;
               graph.SetFillPattern(Gray[ma div attFull],convert_color(r,g,b));
               graph.SetColor(convert_color(r,g,b));
               graph.SetLineStyle(UserBitLn,word((@Gray[ma div attFull])^),NormWidth);
               end
          end
  else with color
       do begin
          Vga256.SetFillColor(convert_color(r,g,b));

(* Le linee che seguono servivano per il driver BGI:
          if not IsBright(color)
          then graph.SetFillStyle(SolidFill,black)
          else graph.SetFillStyle(SolidFill,convert_color(r,g,b));
          graph.SetColor(convert_color(r,g,b));
          graph.SetLineStyle(SolidLn,0,NormWidth);  *)
       end;
end;

procedure sClearDevice;far;
begin
  graph.ClearDevice;
end;

procedure sBar(x1, y1, x2, y2, color: Integer);far;
begin
  graph.SetFillStyle(SolidFill, color);
  graph.Bar(x1,y1,x2,y2);
end;

procedure sline(x1, y1, x2, y2, color: Integer);far;
begin
  graph.SetLineStyle(SolidLn,0,NormWidth);
  graph.SetColor(color);
  graph.line(x1, y1, x2, y2);
end;

procedure sFillPoly(NumPoints: Word; var PolyPoints);far;
begin
  graph.FillPoly(NumPoints,PolyPoints);
end;

function DetectVGA256 : integer;far;
{ Detects VGA or MCGA video cards }
var
  DetectedDriver : integer;
  SuggestedMode  : integer;
begin
  DetectGraph(DetectedDriver, SuggestedMode);
  if (DetectedDriver = VGA) or (DetectedDriver = MCGA) then
    DetectVGA256 := 0        { Default video mode = 0 }
  else
    DetectVGA256 := grError; { Couldn't detect hardware }
end; { DetectVGA256 }

procedure error(s:string);
begin
if graphic=true
then closegraph;
writeln('FATAL ERROR:');
writeln(s);
Halt;
end; { of error }


procedure opengraph(cmode:tmode);
var Driver, Mode, TestDriver,
    ErrCode,i,k: Integer;
    r,g,b:integer;
    Palette: VGA256.PaletteType;
begin
{ inizializza opportunamente il sistema grafico in uso 		}
if cmode in [color256,gray64] (*256*)
then begin
     TestDriver :=
          InstallUserDriver('SVGA256', @DetectVGA256);
     if GraphResult <> grOk
     then error('Error installing VGA256 Driver');
     end;
Driver:=detect;
InitGraph(Driver, Mode, '');
ErrCode := GraphResult;
if ErrCode <> grOk
then error(GraphErrorMsg(ErrCode));
case cmode of
color16:
     begin
     ClearDevice:=sClearDevice;
     Bar:=sBar;
     Line:=sLine;
     FillPoly:=sFillPoly;
     case Driver of
	     Ega: Mode:=egahi;
	     Vga: Mode:=vgamed;
	     else error('Unsupported graphic mode');
        end; { of case }
     SetGraphMode(Mode);
     if graphresult<>GrOk
     then error('Graphics error');
     for i := 0 to 7 do
       with Colors[i] do
         if i = 6
         then SetRGBPalette(20, RedVal, GreenVal, BlueVal)
         else SetRGBPalette(i, RedVal, GreenVal, BlueVal);
     for i := 8 to 15 do
       with Colors[i] do
         SetRGBPalette(i+48, RedVal, GreenVal, BlueVal);
     end;
color256,gray64:
      begin
      if cmode = color256
      then for r:=0 to 7
           do for g:=0 to 7
              do for b:=0 to 3
                 do begin
                      Palette[r*32+g*4+b].r:=r*8;
                      Palette[r*32+g*4+b].g:=g*8;
                      Palette[r*32+g*4+b].b:=b*16;
                    end
      else for i:=0 to 63
           do begin
                Palette[i].r:=i;
                Palette[i].g:=i;
                Palette[i].b:=i;
              end;
      ClearDevice:=Vga256.ClearDevice;
      Bar:=Vga256.Bar;
      Line:=Vga256.Line;
      FillPoly:=Vga256.FillPoly;
      Vga256.InitGraph(buffer256);
      palette256(Palette);
      end;
end; (* of case *)
colormode:=cmode;
graphic:=true;
hmaxx:=getmaxx div 2;
hmaxy:=getmaxy div 2;
multipage:=cmode=color16;
if multipage
then begin
     setactivepage(1);
     ClearDevice;
     end;
pageused:=0;
{ fine inizializzazione del sistema grafico                     }
end; { of OpenGraph }

procedure closegraph;
begin
Graph.CloseGraph;
graphic:=false;
end; { of CloseGraph }

{
uses Crt; (* used by TestPattern *)
procedure TestPattern;
var i,j,m:integer;
begin
  OpenGraph(cmode);
  if multipage
  then SetVisualPage(1);
  case cmode of
    color16: m:=1;
    color256: m:=16;
    gray64: m:=4;
    end;
  for i:=0 to m-1
  do for j:=0 to 15
     do begin
          SetFillStyle(SolidFill,i*16+j);
          graph.Bar(i*10,j*10,i*10+10,j*10+10);
        end;
  while readkey = ''
  do ;
  CloseGraph;
  halt;
end;
}
{$ENDIF}

function Multiply(a,b:byte):byte;assembler;
(* Multiply:=a*b div 256 *)
  asm
  mov al,a
  mul b
  mov al,ah
end;(* of asm *)


function conv(var x,y:real;h:word):word;assembler;
{ la funzione conv realizza la conversione conv:=h-trunc(h*x/y)         }
{ in modo piu' efficiente sfruttando la minore precisione richiesta	}
asm
	les	di,x
	mov	cl,es:[di]
	test	cl,0FFH			{; Ex=0 ?}
	jnz	@continua
@zero:	xor	ax,ax
	jmp	@exit
@continua:
	mov	dx,es:[di+4]
	mov	ax,es:[di+2]		{; dx:ax=(x)}
	mov	ch,dh			{; save sign(x)}
	or	dx,8000H                {; dx:ax=abs(x)}
	les	di,y
	test	byte ptr es:[di],0FFH	{; Ey=0 ?}
	jz	@zero
	sub	cl,es:[di]
	mov	bx,es:[di+4]
	xor	ch,bh
	or	bx,8000H
	cmp	dx,bx
	jb	@do_div
	inc	cl
	shr	dx,1
	rcr	ax,1
@do_div:div	bx
	mul	h
	or	cl,cl
	js	@nega
	jz	@sign_on
	shl	ax,cl
	rcl	dx,cl
        jmp     @sign_on
@nega:  neg	cl
	cmp	cl,20h
	jae	@azzera
	shr	dx,cl
	jmp	@sign_on
@azzera:xor	dx,dx
@sign_on:mov	ax,dx
	or	ch,ch
	js	@exit
	neg	ax
@exit:  add	ax,h
end;
        function Zint(var xi,yi:real;var a,b:tpoint):real;
	var dx,dy:real;
	{ funzione interpolatrice usata dai clippers }
	begin
	dx:=a.x-b.x;
        dy:=a.y-b.y;
	if abs(dx)>abs(dy)
	then Zint:=b.z+(xi-b.x)/dx*(a.z-b.z)
	else Zint:=b.z+(yi-b.y)/dy*(a.z-b.z);
	end; { of Zint }

	function Yint(var xi,zi:real;var a,b:tpoint):real;
	var dx,dz:real;
	{ funzione interpolatrice usata dai clippers }
	begin
	dx:=a.x-b.x;
        dz:=a.z-b.z;
	if abs(dx)>abs(dz)
	then Yint:=b.y+(xi-b.x)/dx*(a.y-b.y)
	else Yint:=b.y+(zi-b.z)/dz*(a.y-b.y);
	end; { of Yint }


	procedure Draw3DClip(var b,a,newa,inters:tproiezione;var found:boolean);

	{ partial 3-D Clipper }

	var d:real;
	    xi,yi,zi:real;
	    status_:byte;

	begin
		  { Si ipotizza che A sia un punto esterno   }
		  { e B un punto interno                     }

	d:=a.rel.x*b.rel.y-a.rel.y*b.rel.x;
	if abs(d)>epsilon
	then if d<0
		then if a.rel.y>=a.rel.x
		     then begin
			     xi:=d/(b.rel.y-a.rel.y+a.rel.x-b.rel.x);
			     yi:=xi;
			     zi:=Zint(xi,yi,a.rel,b.rel);
			     found:=abs(zi)<=xi;
			     status_:=right;
			  end
		     else found:=false
		else if a.rel.y<-a.rel.x
		     then begin
			     xi:=d/(b.rel.y-a.rel.y-a.rel.x+b.rel.x);
			     yi:=-xi;
			     zi:=Zint(xi,yi,a.rel,b.rel);
			     found:=abs(zi)<=xi;
			     status_:=left;
			  end
		     else found:=false
	else found:=false;
	if not found
	then begin
	     d:=a.rel.x*b.rel.z-a.rel.z*b.rel.x;
	     if abs(d)>epsilon
	     then if d<0
		  then if a.rel.z>=a.rel.x
		       then begin
				xi:=d/(b.rel.z-a.rel.z+a.rel.x-b.rel.x);
				zi:=xi;
				yi:=Yint(xi,zi,a.rel,b.rel);
				status_:=up;
				found:=abs(yi)<=xi
			    end
		       else found:=false
		  else  if a.rel.z<-a.rel.x
			then begin
				xi:=d/(b.rel.z-a.rel.z-a.rel.x+b.rel.x);
				zi:=-xi;
				yi:=Yint(xi,zi,a.rel,b.rel);
				status_:=down;
				found:=abs(yi)<=xi
			     end
			else found:=false;
	     end;
	if found
	then begin
		with inters
		do begin
		   xp:=conv(yi,xi,hmaxx);
		   yp:=conv(zi,xi,hmaxy);
		   status:=status_;
		   end;
		newa:=b;
	     end
	{else error('Trap 01 in Draw3DClip');}
	end; { of Draw3DClip }

	procedure Draw3DClip2(var a,b,newa,newb:tproiezione;var found2:boolean);
	var d:real;
	    found:byte;
            i:tpoint;

	{ Si ipotizzano A,B esterni }

	    procedure aggiungi(var pnt:tpoint;status_:byte);
	    begin
	    if status_ and a.status<>0
	    then with newa
			do begin
			    xp:=conv(pnt.y,pnt.x,hmaxx);
			    yp:=conv(pnt.z,pnt.x,hmaxy);
			    status:=status_
			   end
	    else with newb
			do begin
			    xp:=conv(pnt.y,pnt.x,hmaxx);
			    yp:=conv(pnt.z,pnt.x,hmaxy);
			    status:=status_
			   end;
	    inc(found);
	    end; {of aggiungi }



	{ 3-D Clipper }
	begin
	  found:=0;
	  if a.status and b.status=0
	      then begin
		   found:=0;
		   d:=a.rel.x*b.rel.y-a.rel.y*b.rel.x;
		   if abs(d)>epsilon
		   then begin
			if (a.rel.x>a.rel.y)<>(b.rel.x>b.rel.y)
			then begin
			     i.x:=d/(b.rel.y-a.rel.y+a.rel.x-b.rel.x);
			     if i.x>0
			     then begin
				  i.y:=i.x;
				  i.z:=Zint(i.x,i.y,a.rel,b.rel);
				  if abs(i.z)<i.x
				  then aggiungi(i,right);
				  end
			     end;
			if (a.rel.x>-a.rel.y)<>(b.rel.x>-b.rel.y)
			then begin
			     i.x:=d/(b.rel.y-a.rel.y-a.rel.x+b.rel.x);
			     if i.x>0
			     then begin
				  i.y:=-i.x;
				  i.z:=Zint(i.x,i.y,a.rel,b.rel);
				  if abs(i.z)<i.x
				  then aggiungi(i,left);
				  end
			     end;
			end;
		   d:=a.rel.x*b.rel.z-a.rel.z*b.rel.x;
		   if (abs(d)>epsilon)and(found<2)
		   then begin
			if (a.rel.x>a.rel.z)<>(b.rel.x>b.rel.z)
			then begin
			     i.x:=d/(b.rel.z-a.rel.z+a.rel.x-b.rel.x);
			     if i.x>0
			     then begin
				  i.z:=i.x;
				  i.y:=Yint(i.x,i.z,a.rel,b.rel);
				  if abs(i.y)<i.x
				  then aggiungi(i,up);
				  end
			     end;
			if (found<2)and((a.rel.x>-a.rel.z)<>(b.rel.x>-b.rel.z))
			then begin
			     i.x:=d/(b.rel.z-a.rel.z-a.rel.x+b.rel.x);
			     if i.x>0
			     then begin
				  i.z:=-i.x;
				  i.y:=Yint(i.x,i.z,a.rel,b.rel);
				  if abs(i.y)<i.x
				  then aggiungi(i,down);
				  end
			     end;
			end;
		   end;
	      found2:=found=2;
	end; { of Draw3DClip2 }

{$IFDEF WINDOWS}
procedure TGraph3DWindow.visita_spezzata(linep:pline;var lista:vettore_proiezioni);
{$ELSE}
procedure visita_spezzata(linep:pline;var lista:vettore_proiezioni);
{$ENDIF}

var newa,newb:tproiezione;var found:boolean;
    p:pclass;
    a:tproiezione;
    continue:boolean;
    color:integer;

begin
p:=linep^.classe;
continue:=false;
with linep^.colori[1]^.diffuseColor
do color:=convert_color(r,g,b);
with linep^
do   while p<>nil
     do with p^ do
      begin
      if continue
      then begin
	   found:=false;
	   if a.status=0
	   then if lista[vertice].status=0
		then line(a.xp,a.yp,lista[vertice].xp,lista[vertice].yp,color)
		else Draw3DClip(a,lista[vertice],newa,newb,found)
	   else if lista[vertice].status=0
		then Draw3DClip(lista[vertice],a,newa,newb,found)
		else Draw3DClip2(a,lista[vertice],newa,newb,found);
	   if found
	   then line(newa.xp,newa.yp,newb.xp,newb.yp,color)
	   end;
      a:=lista[vertice];
      continue:=cont;
      p:=next
    end
end; { of visita_spezzata }

(* Questa procedura richiede che la procedura FillPoly sia in
   grado di gestire poligoni CONCAVI. E' stata sostituita perche'
   la gestione diretta del video funziona solo per poligoni CONVESSI.

procedure FillSurface(tipo:pmodello;var lista:vettore_proiezioni;
				sup:integer;
				var sup_lum:vettore_sup_luminoso);
const N_Max_Clipped=20;

type Clip=record
		x_,y_:word;
		stato_:byte;
		d_:word;
		p_:pclass;
		next:word;
	  end; { of Clip }

var x,y:integer;
    newa,newb:tproiezione;var found:boolean;
    p:pclass;

    NumPoint:integer;
    Poly:array[0..N_Max_Vertici-1]
	   of record
		  x,y:word
	      end;

    NumClip,ClipLeft,i:word;
    Clipped:array[1..N_Max_Clipped]of Clip;
    Temp:array[1..N_Max_Clipped]of word;
    Visibile:boolean;
    tmp:Clip;ref:boolean;

    a:tproiezione;

	function d(xp,yp:integer;status:byte):word;
	{ questa funzione proietta su una scala lineare il punto Clipped }
	begin
	{ non funziona per i corner !!! }
	case status and $F of
		down:d:=xp;
		right:d:=hmaxx shl 2+hmaxy shl 1+yp;
		up:d:=hmaxx shl 2+hmaxy shl 1-xp;
		left:d:=hmaxx shl 1+hmaxy shl 1-yp;
		end; { of case }

	end; { of d }

  procedure SortClipped(l,r: integer);
  var i,j:integer;
      x:word;
      y:clip;
  begin
  i:=l; j:=r; x:=clipped[(l+r) DIV 2].d_;
  repeat
    while clipped[i].d_<x do inc(i);
    while x<clipped[j].d_ do dec(j);
    if i<=j then
    begin
      y:=clipped[i]; clipped[i]:=clipped[j]; clipped[j]:=y;
      inc(i); dec(j);
    end;
  until i>j;
  if l<j then SortClipped(l,j);
  if i<r then SortClipped(i,r);
  end; { of sort }

	procedure InsertPoly(xp,yp:integer);
	begin
	 with Poly[NumPoint]
	 do begin
	      x:=xp;
	      y:=yp;
	    end;
	 Inc(NumPoint); { controllare Range }
	end; { of InsertPoly }

	procedure InsertClip(xp,yp:integer;status:byte;p:pclass);
	begin
	 Inc(NumClip); { controllare Range }
	 with Clipped[NumClip]
	 do begin
	      x_:=xp;
	      y_:=yp;
	      stato_:=status;
	      d_:=d(xp,yp,status);
	      p_:=p;
	      next:=NumClip;
	    end;
	end; { of InsertClip }

	procedure BoundPoly(j,k:word);
	var i,a,b:integer;
	begin
	{ non funziona per i corner !!! }
	case Clipped[k].stato_ of
	  down:a:=0;
	  right:a:=1;
	  up:a:=2;
	  left:a:=3;
	  end; { of case }
	case Clipped[j].stato_ of
	  down:b:=0;
	  right:b:=1;
	  up:b:=2;
	  left:b:=3;
	  end; { of case }
	if b<a
	then inc(b,4)
	else if b=a
	     then if Clipped[j].d_>Clipped[k].d_
		  then inc(b,4);
	for i:=b-1 downto a
	do case i and 3 of
	   0:InsertPoly(0,hmaxy shl 1);
	   1:InsertPoly(0,0);
	   2:InsertPoly(hmaxx shl 1,0);
	   3:InsertPoly(hmaxx shl 1,hmaxy shl 1)
	   end;
	end; { of BoundPoly }
begin
with tipo^
do begin
		NumClip:=0;
		visibile:=false;
		p:=Superfici[Sup];

		repeat
		{ da ottimizzare .. }
		a:=lista[p^.vertice];
		p:=p^.next;
		while(p<>nil)and(a.status=0)and(lista[p^.vertice].status=0)
		do begin
		    a:=lista[p^.vertice];
		    p:=p^.next;
		    visibile:=true;
		    end;
		if p=nil
		then p:=Superfici[Sup];
		with lista[p^.vertice]
		do if a.status=0
		   then if status=0
			then visibile:=true
			else begin
			     Draw3DClip(a,lista[p^.vertice],newa,newb,found);
                             if found
			     then with newb
				  do InsertClip(xp,yp,status,nil)
			     end
		   else if status=0
			then begin
			     Draw3DClip(lista[p^.vertice],a,newa,newb,found);
                             if found
			     then with newb
				  do InsertClip(xp,yp,status,p);
			     end
			else begin
			     Draw3DClip2(a,lista[p^.vertice],newa,newb,found);
			     if found
			     then begin
				  with newa
				  do InsertClip(xp,yp,status,p);
				  with newb
				  do InsertClip(xp,yp,status,nil)
				  end
			     end;
		until(p=Superfici[Sup]);

		if NumClip=0
		then begin{ no clipping }
		     if visibile
		     then begin
			   NumPoint:=0;
			   p:=Superfici[Sup];{}
			   while p<>nil
			   do begin
				with lista[p^.vertice]
				do InsertPoly(xp,yp);
				p:=p^.next;
			      end;
			   if NumPoint>2
			   then begin
                                SetFillColor(Sup_Lum[Sup]);
                                gFillPoly(NumPoint,Poly);
                                end;
			   end;
		     end
		else begin { clipping on }
		     SortClipped(1,NumClip);
	{ occorre riordinare i punti uscenti ed entranti per i quali	}
	{ le proiezioni coincidono					}
		     i:=1;
		     while(i<NumClip)and(clipped[i].d_=clipped[i+1].d_)
		     do inc(i,2);
		     if i<NumClip
		     then ref:=clipped[i].p_<>nil
		     else ref:=false{(clipped[1].next<clipped[2].next)xor(clipped[1].p_<>nil)};
		     for i:=1 to NumClip-1
		     do if (clipped[i].p_=nil)xor odd(i) xor ref
			then if (clipped[i].d_=clipped[i+1].d_)
			     then begin
				   tmp:=clipped[i];
				   clipped[i]:=clipped[i+1];
				   clipped[i+1]:=tmp
				   end
			     else error('Trap 01 in FillSurface');
	{ fine del riordinamento					}
		     for i:=1 to NumClip
		     do temp[Clipped[i].next]:=i;
		     for i:=1 to NumClip
		     do with Clipped[i]
			do if p_<>nil { entrante }
			   then if next<NumClip
				then next:=temp[next+1]
				else next:=temp[1]
			   else if i<NumClip
				then next:=i+1
				else next:=1;

		     i:=1;ClipLeft:=NumClip;
		     repeat
		     NumPoint:=0;
		     while Clipped[i].next<>0
		     do with Clipped[i]
			do begin
			     InsertPoly(x_,y_);
			     if p_<>nil
			     then begin{ entrante }
				     p:=p_;
				     while (lista[p^.vertice].status=0)
				     do begin
					  with lista[p^.vertice]
					  do InsertPoly(xp,yp);
					  p:=p^.next;
					  if p=nil
					  then p:=Superfici[Sup];
					end;
				     end
			     else BoundPoly(i,next);
			     i:=next;
			     next:=0;
			     Dec(ClipLeft);
			     end;
		      if NumPoint>2
		      then begin
                             SetFillColor(Sup_Lum[Sup]);
                             gFillPoly(NumPoint,Poly);
                           end;
		      if ClipLeft<>0
			then repeat
			     if i<NumClip
			     then inc(i)
			     else i:=1;
			     until Clipped[i].next<>0;
		     until ClipLeft=0;
		     end;
	       end;
end; { of FillSurface }
*)

(* La seguente procedura sostituisce la precedente perche'
   la gestione diretta del video funziona solo per poligoni CONVESSI.*)
{$IFDEF WINDOWS}
procedure TGraph3DWindow.FillSurface(tipo:pmodello;var lista:vettore_proiezioni;
				sup:integer;
				var sup_lum:vettore_sup_luminoso);
{$ELSE}
procedure FillSurface(tipo:pmodello;var lista:vettore_proiezioni;
				sup:integer;
				var sup_lum:vettore_sup_luminoso);
{$ENDIF}
const N_Max_Clipped=20;

type Clip=record
		x_,y_:word;
		stato_:byte;
		d_:word;
		k_:integer;
		next:word;
	  end; { of Clip }

var pc:pPolyCx;
var x,y,k:integer;
    newa,newb:tproiezione;var found:boolean;

    NumPoint:integer;
    Poly:array[0..N_Max_Vertici-1]
	   of record
		  x,y:word
	      end;

    NumClip,ClipLeft,i:word;
    Clipped:array[1..N_Max_Clipped]of Clip;
    Temp:array[1..N_Max_Clipped]of word;
    Visibile:boolean;
    tmp:Clip;ref:boolean;

    a:tproiezione;

	function d(xp,yp:integer;status:byte):word;
	{ questa funzione proietta su una scala lineare il punto Clipped }
	begin
	{ non funziona per i corner !!! }
	case status and $F of
		down:d:=xp;
		right:d:=hmaxx shl 2+hmaxy shl 1+yp;
		up:d:=hmaxx shl 2+hmaxy shl 1-xp;
		left:d:=hmaxx shl 1+hmaxy shl 1-yp;
		end; { of case }

	end; { of d }

  procedure SortClipped(l,r: integer);
  var i,j:integer;
      x:word;
      y:clip;
  begin
  i:=l; j:=r; x:=clipped[(l+r) DIV 2].d_;
  repeat
    while clipped[i].d_<x do inc(i);
    while x<clipped[j].d_ do dec(j);
    if i<=j then
    begin
      y:=clipped[i]; clipped[i]:=clipped[j]; clipped[j]:=y;
      inc(i); dec(j);
    end;
  until i>j;
  if l<j then SortClipped(l,j);
  if i<r then SortClipped(i,r);
  end; { of sort }

	procedure InsertPoly(xp,yp:integer);
	begin
	 with Poly[NumPoint]
	 do begin
	      x:=xp;
	      y:=yp;
	    end;
	 Inc(NumPoint); { controllare Range }
	end; { of InsertPoly }

	procedure InsertClip(xp,yp:integer;status:byte;k:integer);
	begin
	 Inc(NumClip); { controllare Range }
	 with Clipped[NumClip]
	 do begin
	      x_:=xp;
	      y_:=yp;
	      stato_:=status;
	      d_:=d(xp,yp,status);
	      k_:=k;
	      next:=NumClip;
	    end;
	end; { of InsertClip }

	procedure BoundPoly(j,k:word);
	var i,a,b:integer;
	begin
	{ non funziona per i corner !!! }
	case Clipped[k].stato_ of
	  down:a:=0;
	  right:a:=1;
	  up:a:=2;
	  left:a:=3;
	  end; { of case }
	case Clipped[j].stato_ of
	  down:b:=0;
	  right:b:=1;
	  up:b:=2;
	  left:b:=3;
	  end; { of case }
	if b<a
	then inc(b,4)
	else if b=a
	     then if Clipped[j].d_>Clipped[k].d_
		  then inc(b,4);
	for i:=b-1 downto a
	do case i and 3 of
	   0:InsertPoly(0,hmaxy shl 1);
	   1:InsertPoly(0,0);
	   2:InsertPoly(hmaxx shl 1,0);
	   3:InsertPoly(hmaxx shl 1,hmaxy shl 1)
	   end;
	end; { of BoundPoly }
begin
with tipo^
do begin
     pc:=list_PolyCxs[Sup];
     while pc <> nil
     do with pc^
        do begin
		NumClip:=0;
		visibile:=false;
                k:=1;
		repeat
		{ da adattare alla nuova struttura .. }
		a:=lista[vertex[k]];
                inc(k);
		while(k<=n_vertex)and(a.status=0)and(lista[vertex[k]].status=0)
		do begin
		    a:=lista[vertex[k]];
                    inc(k);
		    visibile:=true;
		    end;
		if k>n_vertex
		then k:=1;
		with lista[vertex[k]]
		do if a.status=0
		   then if status=0
			then visibile:=true
			else begin
			     Draw3DClip(a,lista[vertex[k]],newa,newb,found);
                             if found
			     then with newb
				  do InsertClip(xp,yp,status,0)
			     end
		   else if status=0
			then begin
			     Draw3DClip(lista[vertex[k]],a,newa,newb,found);
                             if found
			     then with newb
				  do InsertClip(xp,yp,status,k);
			     end
			else begin
			     Draw3DClip2(a,lista[vertex[k]],newa,newb,found);
			     if found
			     then begin
				  with newa
				  do InsertClip(xp,yp,status,k);
				  with newb
				  do InsertClip(xp,yp,status,0)
				  end
			     end;
		until(k=1);

		if NumClip=0
		then begin{ no clipping }
		     if visibile
		     then begin
			   NumPoint:=0;
                           for k:=1 to n_vertex
			   do with lista[vertex[k]]
		              do InsertPoly(xp,yp);
			   if NumPoint>2
			   then begin
                                SetFillColor(Sup_Lum[Sup]);
                                FillPoly(NumPoint,Poly);
                                end;
			   end;
		     end
		else begin { clipping on }
		     SortClipped(1,NumClip);
	{ occorre riordinare i punti uscenti ed entranti per i quali	}
	{ le proiezioni coincidono					}
		     i:=1;
		     while(i<NumClip)and(clipped[i].d_=clipped[i+1].d_)
		     do inc(i,2);
		     if i<NumClip
		     then ref:=clipped[i].k_<>0
		     else ref:=false{(clipped[1].next<clipped[2].next)xor(clipped[1].p_<>nil)};
		     for i:=1 to NumClip-1
		     do if (clipped[i].k_=0)xor odd(i) xor ref
			then if (clipped[i].d_=clipped[i+1].d_)
			     then begin
				   tmp:=clipped[i];
				   clipped[i]:=clipped[i+1];
				   clipped[i+1]:=tmp
				   end
			     else error('Trap 01 in FillSurface');
	{ fine del riordinamento					}
		     for i:=1 to NumClip
		     do temp[Clipped[i].next]:=i;
		     for i:=1 to NumClip
		     do with Clipped[i]
			do if k_<>0 { entrante }
			   then if next<NumClip
				then next:=temp[next+1]
				else next:=temp[1]
			   else if i<NumClip
				then next:=i+1
				else next:=1;

		     i:=1;ClipLeft:=NumClip;
		     repeat
		     NumPoint:=0;
		     while Clipped[i].next<>0
		     do with Clipped[i]
			do begin
			     InsertPoly(x_,y_);
			     if k_<>0
			     then begin{ entrante }
				     k:=k_;
				     while (lista[vertex[k]].status=0)
				     do begin
					  with lista[vertex[k]]
					  do InsertPoly(xp,yp);
					  inc(k);
					  if k > n_vertex
					  then k:=1;
					end;
				     end
			     else BoundPoly(i,next);
			     i:=next;
			     next:=0;
			     Dec(ClipLeft);
			     end;
		      if NumPoint>2
		      then begin
                             SetFillColor(Sup_Lum[Sup]);
                             FillPoly(NumPoint,Poly);
                           end;
		      if ClipLeft<>0
			then repeat
			     if i<NumClip
			     then inc(i)
			     else i:=1;
			     until Clipped[i].next<>0;
		     until ClipLeft=0;
		     end;
           pc:=next;
           end;
       end;
end; { of FillSurface }

  type cache_rec=record
                   valid:boolean;
                   xr,xl:word;
                   yu,yd:word;
                 end;
(* la variabile <cache> e' dichiarata fuori dalla procedura <sort2>
   per evitare la congestione dello stack *)
  var cache:array[1..n_max_superfici]of cache_rec;

(* le seguenti variabili sono dichiarate fuori dalla procedura <show>
   per evitare la congestione dello stack *)
var    norm:vettore_di_normali;
var    lista:vettore_proiezioni;

{$IFDEF WINDOWS}
procedure TGraph3DWindow.show(p:pnode;xo,yo,zo:real;phi:real;viewangle:real);
{$ELSE}
procedure show(p:pnode;xo,yo,zo:real;phi:real;viewangle:real);
{$ENDIF}

const Max_Obj=30;

var idx2sup:idx_sup; (* loop *)
    idx:integer;  {*loop*}
    sup_lum:vettore_sup_luminoso;
    i:word;
    cl,sl:real;
    n_obj:integer;
    obj_index:integer;
    pos_rel:array[1..MAX_OBJ+1]of tpoint;
    dist:array[1..MAX_OBJ]of real;
    obj:array[1..MAX_OBJ]of pnode;
    cf,sf:real; {cf=cos(phi),sf=sin(phi) }

  {*loop*}
    function IsVertex(v:word;s:pPolyCx):boolean;
    var i:integer;
    (* questa funzione determina se il vertice <v> appartiene a <s> *)
    begin
      i:=1;
      while (i <= s^.n_vertex) and (v <> s^.vertex[i])
      do inc(i);
      IsVertex:=i <= s^.n_vertex;
    end;

  function sup_copre(s1,s2:integer;var mdl:tmodello):boolean;
  { determina se 2 superfici s1 ed s2 dello stesso modello sono tali che	}
  { s1 puo' coprire s2							}

  var p2,p1:pclass;
      AP,OP:tpoint;
      trp:pPolyCx; (*PolyCx*)
      PxP1,QxQ1:tpoint;
      P1next,P2next:word;
      cross:array[1..zMaxPolyCxVertex]of tpoint;
      i:integer;
      dotQxQ1P,dotQxQ1P1,dotPxP1Q,dotPxP1Q1:real;
  begin
{$ifdef stat}
   inc(stat[2]);
{$endif}
  with mdl
  do begin
     trp:=list_PolyCxs[s1];
     while (trp<>nil)
     do with trp^ do
        begin
        p2:=superfici[s2];
        for i:=1 to n_vertex-1
        do crossv(cross[i],Lista[vertex[i]].rel,Lista[vertex[i+1]].rel);
        crossv(cross[n_vertex],Lista[vertex[n_vertex]].rel,Lista[vertex[1]].rel);
        while p2 <> nil
        do begin
           if not IsVertex(p2^.Vertice, trp)
           then begin
                OP:=Lista[p2^.Vertice].rel;
{ la scelta di <vertex[1]> non e' corretta perche' potrebbe appartenere
  all'altra superficie. Modificare la seguente linea }
                diff(AP,OP,Lista[vertex[1]].rel);
{ il confronto con <epsilon> serve a evitare errori di arrotondamento }
	        if (dot(norm[s1],AP) < -epsilon) (* condizione D *)
                then begin
                     i:=1;
                     while (i <= n_vertex) and (dot(cross[i],OP) < 0)
                     do inc(i);
                     if i > n_vertex
                     then begin
{$ifdef stat}
                             inc(stat[3]);
{$endif}
                             sup_copre:=true;
                             exit;
                          end;
                     end;
                end;
	   p2:=p2^.next;
           end;
        trp:=trp^.next;
        end;
     end;
  with mdl
  do begin
     trp:=list_PolyCxs[s2];
     while (trp<>nil)
     do with trp^ do
        begin
        p1:=superfici[s1];
        for i:=1 to n_vertex-1
        do crossv(cross[i],Lista[vertex[i]].rel,Lista[vertex[i+1]].rel);
        crossv(cross[n_vertex],Lista[vertex[n_vertex]].rel,Lista[vertex[1]].rel);
        while p1 <> nil
        do begin
           if not IsVertex(p1^.Vertice, trp)
           then begin
                OP:=Lista[p1^.Vertice].rel;
{ la scelta di <vertex[1]> non e' corretta perche' potrebbe appartenere
  all'altra superficie. Modificare la seguente linea }
                diff(AP,OP,Lista[vertex[1]].rel);
{ il confronto con <epsilon> serve a evitare errori di arrotondamento }
	        if (dot(norm[s2],AP) > epsilon) (* condizione D *)
                then begin
                     i:=1;
                     while (i <= n_vertex) and (dot(cross[i],OP) < 0)
                     do inc(i);
                     if i > n_vertex
                     then begin
{$ifdef stat}
                             inc(stat[4]);
{$endif}
                             sup_copre:=true;
                             exit;
                          end;
                     end;
                end;
	   p1:=p1^.next;
           end;
        trp:=trp^.next;
        end;
     end;
  with mdl
  do begin
       p1:=superfici[s1];
       while (p1<>nil)
       do begin
            if p1^.next=nil
            then P1next:=superfici[s1]^.vertice
            else P1next:=p1^.next^.vertice;
            crossv(PxP1,Lista[p1^.vertice].rel,Lista[P1next].rel);
            p2:=superfici[s2];
            while (p2<>nil)
            do begin
               if p2^.next=nil
               then P2next:=superfici[s2]^.vertice
               else P2next:=p2^.next^.vertice;
               if (p1^.vertice <> p2^.vertice)and(P1next <> p2^.vertice)
                  and (p1^.vertice <> P2next)and(P1next <> P2next)
               then begin
                    dotPxP1Q:=dot(PxP1,Lista[p2^.vertice].rel);
                    dotPxP1Q1:=dot(PxP1,Lista[P2next].rel);
                    if  (dotPxP1Q > 0)xor
                        (dotPxP1Q1 > 0)
                    then begin
                         crossv(QxQ1,Lista[p2^.vertice].rel,Lista[P2next].rel);
                         dotQxQ1P:=dot(QxQ1,Lista[p1^.vertice].rel);
                         dotQxQ1P1:=dot(QxQ1,Lista[P1next].rel);
                         if (dotQxQ1P > 0)xor
                            (dotQxQ1P1 > 0)
                         then if abs(dotQxQ1P-dotQxQ1P1) > abs(dotPxP1Q-dotPxP1Q1)
                              then begin
{$ifdef stat}
                                     inc(stat[5]);
{$endif}
                                     sup_copre:=true;
                                     exit;
                                   end;
                         end;
                    end;
                 p2:=p2^.next;
               end;
            p1:=p1^.next;
          end;
     end;
  sup_copre:=false;
  end; { of sup_copre }


  procedure sort2(var mdl:tmodello);
  { effettua un riordinamento dinamico delle superfici }
  var first,i,last:integer;
      att:integer;

     procedure swap(a,b:integer);
     var t:integer;
         tc:cache_rec;
     begin
       t:=idx2sup[a];
       idx2sup[a]:=idx2sup[b];
       idx2sup[b]:=t;
       tc:=cache[a];
       cache[a]:=cache[b];
       cache[b]:=tc;
     end;

   function sup_copre2(s1,s2:integer;var mdl:tmodello):boolean;
   { determina se 2 superfici s1 ed s2 dello stesso modello sono tali che	}
   { s1 puo' coprire s2							}

   var xr1,xl1,yu1,yd1,
       xr2,xl2,yu2,yd2:word;
       t:plist_sup;

       procedure BoundRect(ss:integer;var xr,xl,yu,yd:word);
       (* Determina il rettangolo limite *)
       var p:pclass;

           procedure analizza2(var a:tproiezione);
           begin
           with a
           do begin
              if status and right = right
              then xp:=0
              else if status and left = left
                   then xp:=2*hmaxx
                   else if status <> 0
                        then xp:=conv(rel.y,rel.x,hmaxx);
              if status and up = up
              then yp:=0
              else if status and down = down
                   then yp:=2*hmaxy
                   else if status <> 0
                        then yp:=conv(rel.z,rel.x,hmaxy);
              end;
           end;
       begin
         if cache[ss].valid
         then begin
              xr:=cache[ss].xr;
              xl:=cache[ss].xl;
              yu:=cache[ss].yu;
              yd:=cache[ss].yd;
              end
         else begin
              p:=mdl.superfici[idx2sup[ss]];
              analizza2(Lista[p^.vertice]);
              xl:=Lista[p^.vertice].xp;
              xr:=xl;
              yu:=Lista[p^.vertice].yp;
              yd:=yu;
              p:=p^.next;
              while (p<>nil)
              do begin
                 analizza2(Lista[p^.vertice]);
                 if Lista[p^.vertice].xp > xr
                 then xr:=Lista[p^.vertice].xp
                 else if Lista[p^.vertice].xp < xl
                      then xl:=Lista[p^.vertice].xp;
                 if Lista[p^.vertice].yp > yd
                 then yd:=Lista[p^.vertice].yp
                 else if Lista[p^.vertice].yp < yu
                      then yu:=Lista[p^.vertice].yp;
                 p:=p^.next;
	         end;
              cache[ss].valid:=true;
              cache[ss].xr:=xr;
              cache[ss].xl:=xl;
              cache[ss].yu:=yu;
              cache[ss].yd:=yd;
              end;
       end;

   begin
{$ifdef stat}
   inc(stat[0]);
{$endif}
   t:=mdl.sup_coperte^[idx2sup[s1]];
   while (t<>nil) and (t^.sup<>idx2sup[s2])
   do t:=t^.next;
   if t=nil
   then sup_copre2:=false (* l'analisi statica ha gia' determinato la non copertura *)
   else begin
{$ifdef stat}
        inc(stat[1]);
{$endif}
        BoundRect(s1,xr1,xl1,yu1,yd1);
        BoundRect(s2,xr2,xl2,yu2,yd2);
        if (xr1 < xl2) or (xr2 < xl1) or (yd1 < yu2) or (yd2 < yu1)
        then sup_copre2:=false (* i rettangoli limite non si intersecano *)
        else sup_copre2:=sup_copre(idx2sup[s1],idx2sup[s2],mdl);
        end;
   end; { of sup_copre2 }


  begin
    if idx = 0
    then exit;
    last:=idx;
    for i:=1 to idx
    do begin
       cache[i].valid:=false;
       if idx2sup[i] <= mdl._last
       then last:=i;
       end;
  { In presenza di relazioni	                                }
  { mutuamente ricorsive fra le superfici, il risultato non e'	}
  { corretto ( potrebbero verificarsi LOOPs )			}
    first:=1;att:=0;
    while (first <> last) and (att < idx)
    do begin
       i:=first+1;
       while (i<=last) and not sup_copre2(first,i,mdl)
       do inc(i);
       if i<=last
       then begin
              if i<> first+1
              then swap(first,first+1);
	      swap(i,first);
              inc(att);
	    end
       else begin
	       inc(first);att:=0;
	    end;
       end;
{$ifdef debug}
    if att = idx
    then error('runtime loop detected');
{$endif}
  end; { of sort_sup }

  procedure sort(l,r: integer);
  var i,j:integer;
      x,y:real;
      p:pnode;
      t:tpoint;
  begin
  i:=l; j:=r; x:=dist[(l+r) DIV 2];
  repeat
    while dist[i]>x do i:=i+1;
    while x>dist[j] do j:=j-1;
    if i<=j then
    begin
      y:=dist[i]; dist[i]:=dist[j]; dist[j]:=y;
      p:=obj[i]; obj[i]:=obj[j]; obj[j]:=p;
      t:=pos_rel[i]; pos_rel[i]:=pos_rel[j]; pos_rel[j]:=t;
      i:=i+1; j:=j-1;
    end;
  until i>j;
  if l<j then sort(l,j);
  if i<r then sort(i,r);
  end; { of sort }

  procedure analizza(var a:tproiezione);

  {							}
  {     Questa procedura determina a partire dalle	}
  { 	coordinate relative i seguenti parametri:	}
  {							}
  {	xp,yp:proiezioni sul piano virtuale             }
  {	status: informazione di visibilita' i punti	}
  {	   	   	+-----+-----+-----+-----+-----+	}
  {		status= !retro!right!left !down !up   !	}
  {			+-----+-----+-----+-----+-----+	}
  {							}
  {		retro:il punto si trova alle spalle     }
  {			dell'osservatore		}
  {	Nota: status=0 <==> visibilita' del punto	}

  begin
	with a do
	if rel.x>d_min
	then begin
		 if rel.z>=rel.x
		 then status:=up
		 else if rel.z<-rel.x
		      then status:=down
		      else status:=0;
		 if rel.y>=rel.x
		 then status:=status or right
		 else if rel.y<-rel.x
		      then status:=status or left;
		 if status=0
		 then begin {il punto e' visibile }
			 xp:=conv(rel.y,rel.x,hmaxx);
			 yp:=conv(rel.z,rel.x,hmaxy)
		      end;
		 end
	      else begin
		   status:=retro;
		   if rel.z>=0
		   then status:=status or up
		   else status:=status or down;
		   if rel.y>=0
		   then status:=status or right
		   else status:=status or left;
		   end;
  end;

  procedure prepara_sup(var pobj:pnode);
  var i:integer;
(* Questa procedura determina quali superfici dell'oggetto sono visibili.
   La verifica e' fatta sulla posizione dei vertici della superficie e
   sulla direzione della normale.
   I risultati di questa procedura influiscono in particolare sull'efficienza
   della routine <sort2>. *)
     function peso(status:byte):integer;
     var w:integer;
     begin
     if status and right = right
     then w:=1
     else w:=0;
     if status and left = left
     then inc(w);
     if status and up = up
     then inc(w);
     if status and down = down
     then inc(w);
     peso:=w;
     end;

     function sup_visibile(p:pclass):boolean;
     var visibile,pvisibile:boolean;
         ostatus,first:byte;
     begin
       pvisibile:=false;
       visibile:=false;
       first:=lista[p^.vertice].status;
       ostatus:=first;
       while p<>nil
       do with lista[p^.vertice],pobj^
       do begin
          if (status = 0) or (peso(ostatus xor status) > peso(ostatus))
          then pvisibile:=true;
          ostatus:=status;
          if not visibile and(status and retro=0)and
             (dot(norm[i],rel)<0)
          then visibile:=true;
          p:=p^.next;
          end;
       if peso(ostatus xor first) > peso(ostatus)
       then pvisibile:=true;
       sup_visibile:=visibile and pvisibile;
     end;

     procedure adjust(var a,b,c:real);
     var att,max:real;
     begin
       if a > b
       then if a > c
            then max:=a
            else max:=c
       else if b > c
            then max:=b
            else max:=c;
       if max > 255
       then begin
            att:=255.9/max;
            a:=a*att;
            b:=b*att;
            c:=c*att;
            end;
     end;

     procedure colora(p:pclass;material:pmaterial);
     var abs_norm,r,temp:tpoint;
         diff,ref,red,blue,green:real;
         light:pLight;
         is_center_def:boolean;center:tpoint;q:pclass;
     begin
       light:=currentLight;
       with pobj^,state,tipo^.normali[i]
       do begin
            abs_norm.x:=x*cs-y*sn;
            abs_norm.y:=y*cs+x*sn;
            abs_norm.z:=z;
          end;
       with material^.emissiveColor
       do begin
            red:=r;
            green:=g;
            blue:=b;
          end;
       is_center_def:=false; (* indica che il baricentro non e' stato calcolato *)
       while light <> nil
       do begin
            diff:=-dot(light^.dir,abs_norm); (* NON SPOSTARE *)
            if  isBright(material^.specularColor)and(diff>0)
            then begin
                 scalar(temp,abs_norm,2*diff);
                 sum(r,light^.dir,temp);
                 if not is_center_def
                 then begin
                      q:=p;
                      center.x:=0;
                      center.y:=0;
                      center.z:=0;
                      while q <> nil
                      do begin
                         with lista[p^.vertice]
                         do begin
                            center.x:=center.x+rel.x;
                            center.y:=center.y+rel.y;
                            center.z:=center.z+rel.z;
                            end;
                         q:=q^.next;
                         end;
                      is_center_def:=true;
                      versore(center);
                      end;
                 ref:=dot(light^.dir,center);
                 if ref > 0
                 then with light^,material^.specularColor
                      do begin
                           red:=red+Multiply(r,color.r)*ref;
                           green:=green+Multiply(g,color.g)*ref;
                           blue:=blue+Multiply(b,color.b)*ref;
                         end;
                 end;
            if  isBright(material^.diffuseColor)and(diff > 0)
            then with light^,material^.diffuseColor
                 do begin
                      red:=red+Multiply(r,color.r)*diff;
                      green:=green+Multiply(g,color.g)*diff;
                      blue:=blue+Multiply(b,color.b)*diff;
            end;
          light:=light^.next;
          end;
       adjust(red,green,blue);
       sup_lum[i].r:=trunc(red);
       sup_lum[i].g:=trunc(green);
       sup_lum[i].b:=trunc(blue);
     end;

  begin
    idx:=0;
    with pobj^,tipo^,state do
    for i:=1 to n_superfici
    do begin
       with normali[i]
       do begin
          norm[i].x:=x*(cs*cf+sn*sf) + y*(sf*cs-sn*cf);
          norm[i].y:=y*(cs*cf+sn*sf) - x*(sf*cs-sn*cf);
          norm[i].z:=z;
          end;
       if sup_visibile(superfici[i])
       then begin
              inc(idx);
              idx2sup[idx]:=i;
              colora(superfici[i],colori[i]);
  	    end;
       end;
  end;

  procedure store_obj(p:pnode);
  var dx,dy:real;
      xr,yr,zr:real;
      estensione:real;
  begin
  while p<>nil
  do begin
     case p^.typenode of
       lineset, faceset:
             with p^,state
             do if attivo
                then begin
		          dx:=posizione.x-xo;
		          dy:=posizione.y-yo;
		          xr:=dx*cf+dy*sf;
		          yr:=dy*cf-dx*sf;
		          zr:=posizione.z-zo;
		          { salva xr,yr,zr per evitare di ricalcolarli dopo }
		          with pos_rel[n_obj+1]
		          do begin
		             x:=xr;
		             y:=yr;
		             z:=zr;
		             end;
	                  cl:=(cs*cf+sn*sf)*scala;
		          sl:=(sf*cs-sn*cf)*scala;
(* qui l'identificatore <tipo> e' usato con lo stesso significato di <line> *)
                          with tipo^.PointArrayp^
                          do begin
                             xr:=xr+middle.x*cl+middle.y*sl;
                             yr:=yr+middle.y*cl-middle.x*sl;
                             zr:=zr+middle.z*scala;
                             end;
                          estensione:=scala*tipo^.extension;
		          if (xr>-estensione)and(abs(yr)<xr+estensione*1.414)
		          and(abs(zr)<xr+estensione*1.414)
		          then      { l'oggetto e' probabilmente visibile }
		                begin
		                 inc(n_obj);
		                 if n_obj>max_obj
		                 then error('Max_Obj is too small');
		                 dist[n_obj]:=sqr(xr)+sqr(yr)+sqr(zr);
		                 obj[n_obj]:=p;
		                end;
	             end;
         structure:if p^.attivo then store_obj(p^.child);
       end; (* of case *)
     p:=p^.next;
     end;
  end;


begin
if currentLight = nil
then begin
     new(currentLight);
     with currentLight^
     do with color,dir
        do begin
             r:=255;g:=255;b:=255;
             x:=0;y:=0;z:=-1;
             next:=nil;
           end;
     end;

{$IFNDEF WINDOWS}
if multipage
then begin
     pageused:=pageused xor 1; { switch page }
     SetActivePage(pageused);
     end;
{$ENDIF}
if (sky.r=0)and(sky.g=0)and(sky.b=0)and(ground.r=0)and(ground.g=0)and(ground.b=0)
then begin
     Cleardevice;
     {$ifdef hline}
     line(0,hmaxy,hmaxx * 2,hmaxy,convert_color(255,255,255));
     {$endif}
     end
else begin
     with sky
     do Bar(0,0,hmaxx*2,hmaxy, convert_color(r,g,b));
     {$ifdef hline}
     line(0,hmaxy,hmaxx * 2,hmaxy,
          convert_color(127+sky.r div 2,127+sky.g div 2,127+sky.b div 2));
     {$endif}
     with ground
     do Bar(0,hmaxy+1,hmaxx*2,hmaxy*2, convert_color(r,g,b));
     end;
n_obj:=0;
cf:=cos(phi);
sf:=sin(phi);
store_obj(p);
if n_obj>1
then sort(1,n_obj);
for obj_index:=1 to n_obj
do with obj[obj_index]^,state
   do begin
      if attivo
      then begin
(* qui l'identificatore <tipo> e' usato con lo stesso significato di <line> *)
		with tipo^.PointArrayp^
		do begin
			cl:=(cs*cf+sn*sf)*scala;
			sl:=(sf*cs-sn*cf)*scala;
			for i:=0 to n_vertici-1
			do with lista[i],vertici[i]
			   do begin
			      rel.x:=x*cl+y*sl+pos_rel[obj_index].x;
			      rel.y:=y*cl-x*sl+pos_rel[obj_index].y;
			      rel.z:=z*scala+pos_rel[obj_index].z;
			      analizza(lista[i]);
			      end;
                        case obj[obj_index]^.typenode of
                        faceset:
			     begin
                             prepara_sup(obj[obj_index]);
                             if idx <> 0
                             then begin
                                  if tipo^.loop
                                  then sort2(obj[obj_index]^.tipo^); {*loop*}
                                  for i:=1 to idx
                                  do FillSurface(obj[obj_index]^.tipo,lista,idx2sup[i],sup_lum);
                                  end;
                             end;
                        lineset:visita_spezzata(obj[obj_index]^.line,lista);
                        end; (* of case *)
		end;
	   end;
   end;
{$IFDEF WINDOWS}
transferImage
{$ELSE}
if multipage
then SetVisualPage(pageused)
else Vga256.transferBuffer;
{$ENDIF}
end; { of show }

{$IFNDEF WINDOWS}
begin
graphic:=false;
new(buffer256);
{$ifdef stat}
for istat:=0 to max_stat
do stat[istat]:=0;
{$endif}
{$ENDIF}
end.
