unit vga256;
(* programma per la gestione diretta del video 320x200 a 256 colori *)
(* sono possibili due modi di funzionamento: con o senza buffer:
   per usare il buffer occorre inserire all'inizio del programma la
   procedura new(buffer) e poi utilizzare la procedura TransferBuffer
   per trasferire il buffer sul video

   ATTENZIONE: le routine incluse in questo file non verificano la
               correttezza dei dati che vengono passati!!
*)
interface
type colors     = record
                    r,g,b:byte;
                  end;
    PaletteType = array[0..255]of colors;
    BufferType  = array[0..199,0..319]of byte;
    pBufferType = ^BufferType;
    tPoint      = record
                    x,y:integer;
                  end;

var VideoFrame: BufferType absolute $A000:$0000;

procedure InitGraph(buf:pBufferType);
procedure CloseGraph;
procedure Palette256(var Palette);
procedure TransferBuffer;
procedure ClearDevice;
procedure Bar(x1,y1,x2,y2,col:integer);
procedure PutPixel(x,y,col:integer);
procedure Line(x1,y1,x2,y2,col:integer);
procedure FillPoly(NumPoints: word;var PolyPoints);
procedure SetFillColor(col:integer);

implementation
var fillColor:integer;
    buffer:pBufferType;

procedure InitGraph(buf:pBufferType);
begin
  buffer:=buf;
  asm
    mov ax,013h
    int 10h
    les di,buffer
    xor ax,ax
    mov cx,32000
    cld
    rep stosw
  end;
end;

procedure CloseGraph;assembler;
asm
  mov ax,03h
  int 10h
end;

procedure Palette256(var Palette);assembler;
asm
  les dx,dword ptr Palette
  mov ax,1012h
  mov bx,0
  mov cx,100h
  int 10h
end;

procedure TransferBuffer;assembler;
asm
  push ds
  lds si,buffer
  mov ax, seg word ptr VideoFrame
  mov es,ax
  lea di, VideoFrame
  mov cx,32000
  cld
  rep movsw
  pop ds
end;


procedure ClearDevice;assembler;
asm
  les di,buffer
  xor ax,ax
  mov cx,32000
  cld
  rep stosw
end;

procedure SetFillColor(col:integer);
begin
  FillColor:=col;
end;

procedure Bar(x1,y1,x2,y2,col:integer);assembler;
asm
  les di,buffer
  mov cx,y2
  mov ax,y1
  sub cx,ax
  jns @checkx
  mov ax,y2
  neg cx
@checkx:
  inc cx
  shl ax,6
  add di,ax
  shl ax,2
  add ax,x1
  add di,ax
  cld
  mov al,byte ptr col
@nexty:
  push cx
  push di
  mov cx,x2
  sub cx,x1
  inc cx
  rep stosb
  pop di
  add di,320
  pop cx
  loop @nexty
end;

procedure PutPixel(x,y,col:integer);assembler;
asm
  les di,buffer
  mov ax,y
  shl ax,6
  add di,ax
  shl ax,2
  add ax,x
  add di,ax
  mov al,byte ptr col
  mov es:[di],al
end;

procedure Line(x1,y1,x2,y2,col:integer);assembler;
var absx,absy,deltax,deltay,mm:integer;
asm
(* set-up variables needed to draw the line *)
  les     di,buffer
  mov     ax,y1
  shl     ax,6
  add     di,ax
  shl     ax,2
  add     ax,x1
  add     di,ax
  mov     ax,x1
  mov     cx,x2
  SUB     CX,AX
  JGE     @deltax1
  NEG     CX
  MOV     ax,-1
  JMP     @checky
@deltax1:
  MOV     ax,1
@checky:
  mov     bx,y1
  mov     dx,y2
  SUB     DX,BX
  JGE     @deltay1
  NEG     DX
  MOV     bx,-320
  JMP     @skip1
@deltay1:
  MOV     bx,320
@skip1:
  CMP     DX,CX
  JGE     @drawy
  XCHG    DX,CX
  XCHG    AX,BX

@drawy:
  MOV     [DELTAX],AX
  MOV     [DELTAY],BX
  MOV     [ABSX],CX
  MOV     [ABSY],DX
  mov     bl,byte ptr col
  MOV     AX,[ABSY]
  MOV     CX,AX
  SHR     AX,1
  MOV     [MM],AX
@loopy:
  MOV     ES:[DI],BL
  JCXZ    @exit
  DEC     CX
  MOV     DX,[DELTAY]
  MOV     AX,[ABSX]
  ADD     [MM],AX
  MOV     AX,[MM]
  CMP     AX,[ABSY]
  JL      @skipy
  ADD     DX,[DELTAX]
  MOV     AX,[ABSY]
  SUB     [MM],AX
@skipy:
  ADD     DI,DX
  JMP     @loopy
@exit:
end;

procedure HLine(x1,x2,y:integer;col:byte);assembler;
(* disegna una linea orizzontale fra (x1,y) e (x2,y) di colore col *)
asm
  mov bx,x1
  mov cx,x2
  cmp bx,cx
  jbe @no_exch
  xchg bx,cx
@no_exch:
  sub cx,bx
  inc cx
  les di,buffer
  mov ax,y
  shl ax,6
  add di,ax
  shl ax,2
  add ax,bx
  add di,ax
  mov al,col
  cld
  rep stosb
end;

procedure FillPoly(NumPoints: word;var PolyPoints);assembler;
var d1y,dd2y,dd1y,d2y,col:byte;
    ymax,d1x,_ss,dd2x,dd1x,y2,x2,nn,y1,x1,mm,y0,x0,d2x,s,n,m,j0,j2,np:word;

{
 questa procedura riempe poligoni CONVESSI con il colore <col>

 np e' il numero di vertici del poligono

 Poly e' un array di record di due word ciascuno:
        Poly:array[1..np]of record
			   x,y:word;
			 end
}
asm
	push	ds
        mov     al,byte ptr FillColor
        mov     col,al
        les     di,buffer
	mov	ax,NumPoints
	lds     si,PolyPoints		{ load Poly address }
	cmp	ax,3			{ check that np >= 3 }
	jae	@proceed_fill
	jmp	@exit_fillp
(* la subroutine hline traccia una linea verticale da (bx,ax) a (cx,ax) *)
@hline:
        cmp     bx,cx
        jbe     @no_exch
        xchg    bx,cx
@no_exch:
        sub     cx,bx
        inc     cx
        shl     ax,6
        push    di
        add     di,ax
        shl     ax,2
        add     ax,bx
        add     di,ax
        mov     al,col
        cld
        rep     stosb
        pop     di
        retn
@_fill:
(* first check that x0,x1,x2 are in their ranges
   note that y0,y1,y2 are checked by the caller fill *)
	cmp	x0,320                  { check x0 }
	jae	@exit_fill
	cmp	x1,320                  { check x1 }
	jae	@exit_fill
	cmp	x2,320                  { check x2 }
	jnae	@proceed
@exit_fill:
        retn
(* now proceed with filling *)
@proceed:
	mov	ax,y1
	mov	si,ax		        { yy = y1  }
	sub	ax,y0
	jns     @1
	neg	ax
@1:
	xchg	ax,dx
	mov	ax,y1
	sub	ax,y2
	jns	@2
	neg	ax
@2:
	cmp	dx,ax		{ dx = abs(y0-y1), ax = abs(y2-y1) }
	jle	@3@1
	mov	ax,y2
	jmp	@3@2
@3@1:
	mov	ax,y0
@3@2:
	mov	y1,ax
	jmp	@3@9
@3@3:
	mov	ax,n
	add	s,ax
	jns	@3@4
	mov	ax,m
	add	s,ax
	mov	cx,x2
	mov	ax,d2x
	add	x2,ax
	mov	al,byte ptr d2y
	jmp	@3@5
@3@4:
	mov	cx,x2
	mov	ax,d1x
	add	x2,ax
	mov	al,byte ptr d1y
@3@5:
	or	al,al
	jne	@3@3
@3@6:
	mov	ax,nn
	add	_ss,ax
	jns	@3@7
	mov	ax,mm
	add	_ss,ax
	mov	bx,x0
	mov	ax,dd2x
	add	x0,ax
	mov	al,byte ptr dd2y
	jmp	@3@8
@3@7:
	mov	bx,x0
	mov	ax,dd1x
	add	x0,ax
	mov	al,byte ptr dd1y
@3@8:
	or	al,al
	jne	@3@6
	mov	ax,si		{ load yy }

	call	@hline
	inc	si		{ inc(yy) }
@3@9:
	cmp	si,y1
	jge	@exit_loop
	jmp	@3@3
@exit_loop:
        retn
@proceed_fill:
        dec     ax
	shl	ax,2
	mov	np,ax			{ np:=np*sizeof(dword) }

	xor	cx,cx			{ ora cerca ymin }
	mov	dx,ds:[si][2]        	{ ymin:=Poly[0].y }
	mov	bx,4                    { i:=1          }
	jmp	@find@min
@min:
	mov	ax,ds:[si][bx][2]
	cmp	ax,dx
	jae	@next@min
	mov	cx,bx
	mov	dx,ax
@next@min:
	add	bx,4
@find@min:
	cmp	bx,np
	jbe	@min
	cmp	dx,200                  { check ymin         }
	jnc	@exit_fillp
	mov	j0,cx			{ cx:=imin, dx:=ymin }
	mov	j2,cx
	mov	bx,cx
	mov	ax,ds:[si][bx]
	mov	x1,ax
	mov	ax,ds:[si][bx+2]
	mov	y1,ax
	mov	y0,ax
	mov	y2,ax
	xor	bx,bx
	mov	ymax,bx		{ ymax:=0                    }
	jmp	@find@max
@max:
	mov	ax,ds:[si][bx+2]
	cmp	ax,ymax
	jbe	@next@max
	mov	ax,ds:[si][bx+2]
	mov	ymax,ax
@next@max:
	add	bx,4
@find@max:
	cmp	bx,np
	jle	@max
	cmp	ymax,200                  { check ymax       }
	jc	@while
        jmp     @exit_fillp
@while:
	mov	ax,y1		{ while y1<>ymax             }
	cmp	ax,ymax         { do                         }
	jne	@do	   	{                            }
	jmp	@last@line
@do:
	mov	ax,y1
	cmp	ax,y0
	je	@newj0
	jmp	@checkj2
@newj0:
	mov	bx,j0		{ bx:=j0                     }
	mov	cx,ds:[si][bx]
@decj0:
	mov	ax,ds:[si][bx]
	mov	x1,ax
	or	bx,bx
	jle	@j0@np
	sub	bx,4
	jmp	@j0@cont
@j0@np:
	mov	bx,np
@j0@cont:
	mov	ax,ds:[si][bx][2]
	cmp	ax,y0
	jbe	@decj0
	mov	j0,bx
	cmp	x1,cx
	je	@j0@noline
	mov	ax,y1
	mov 	bx,cx
	mov	cx,x1
	call	@hline
@j0@noline:
	mov	bx,j0
	mov	ax,ds:[si][bx]
	mov	x0,ax
	mov	ax,ds:[si][bx+2]
	mov	y0,ax
	mov	ax,x0
	sub	ax,x1
	js	@dd1x@neg
	mov	dd1x,1
	jmp	@dd1x@cont
@dd1x@neg:
	mov	dd1x,-1
	neg	ax
@dd1x@cont:            		{ dd1x=sign(x0-x1)           }
	mov	cx,ax		{ cx = abs(x0-x1)            }
	mov	ax,y0
	sub	ax,y1
	jns	@mm
	neg	ax
@mm:
	mov	mm,ax		{ mm = abs(y0-y1)            }
	jne	@dd1y@0
	mov	al,1
	jmp	@dd1y@cont
@dd1y@0:
	xor	al,al
@dd1y@cont:
	mov	dd1y,al
	mov	dx,mm
	sub	dx,cx			{ dx=nn              }
	jge	@nn@pos
	mov	mm,cx
	mov	ax,dd1x
	mov	dd2x,ax
	mov	dd2y,1
	jmp	@nn@cont
@nn@pos:
	neg	dx
	mov	dd2x,0
	mov	al,dd1y
	mov	dd2y,al
@nn@cont:
	mov	nn,dx
	mov	ax,mm
	sar	ax,1
	mov	_ss,ax
	mov	ax,x1
	mov	x0,ax
@checkj2:
	mov	ax,y1
	cmp	ax,y2
	je	@newj2
	jmp	@call@fill
@newj2:
	mov	bx,j2		{ bx:=j2                     }
	mov	cx,ds:[si][bx]
@incj2:
	mov	ax,ds:[si][bx]
	mov	x1,ax
	cmp	bx,np
	jge	@j2@0
	add	bx,4
	jmp	@j2@cont
@j2@0:
	xor	bx,bx
@j2@cont:
	mov	ax,ds:[si][bx][2]
	cmp	ax,y2
	jbe	@incj2
	mov	j2,bx
	cmp	x1,cx
	je	@j2@noline
	mov	ax,y1
	mov 	bx,cx
	mov	cx,x1
	call	@hline
@j2@noline:
	mov	bx,j2
	mov	ax,ds:[si][bx]
	mov	x2,ax
	mov	ax,ds:[si][bx][2]
	mov	y2,ax
	mov	ax,x2
	sub	ax,x1
	js	@d1x@neg
	mov	d1x,1
	jmp	@d1x@cont
@d1x@neg:
	mov	d1x,-1
	neg	ax
@d1x@cont:			{ d1x=sign(x2-x1)            }
	mov	cx,ax		{ cx =abs(x2-x1)             }
	mov	ax,y2
	sub	ax,y1
	jns	@m
	neg	ax
@m:
	mov	m,ax
	jne	@d1y@0
	mov	al,1
	jmp	@d1y@cont
@d1y@0:
	xor	al,al
@d1y@cont:
	mov	d1y,al
	mov	dx,m
	sub	dx,cx
        jge	@n@pos
	mov	m,cx
	mov	ax,d1x
	mov	d2x,ax
	mov	d2y,1
	jmp	@n@cont
@n@pos:
	neg	dx
	mov	d2x,0
	mov	al,d1y
	mov	d2y,al
@n@cont:
	mov	n,dx
	mov	ax,m
	sar	ax,1
	mov	s,ax
	mov	ax,x1
	mov	x2,ax
@call@fill:
	push	si
	call	@_fill
	pop	si
	jmp	@while
@last@line:
	mov	ax,y1
	mov	bx,j2
	mov	cx,ds:[si][bx]
	mov	bx,j0
	mov	bx,ds:[si][bx]
	call	@hline
@exit_fillp:
        pop     ds
end;

{ ESEMPIO DI USO DELLA UNIT 320X200
uses crt,320x200;
const
 Esa: array[1..6] of tPoint = ((X: 100; Y: 50),(X:200; Y: 50),
   (X: 250; Y: 100), (X: 200; Y: 150), (X: 100; Y: 150), (X: 50; Y: 100));
var i,j,r,g,b:word;
    Palette:PaletteType;
begin
buffer:=@VideoFrame; (* indirizza le procedure grafiche direttamente sul video *)
InitGraph;
for r:=0 to 7
do for g:=0 to 7
   do for b:=0 to 3
      do begin
           Palette[r*32+g*4+b].r:=r*8;
           Palette[r*32+g*4+b].g:=g*8;
           Palette[r*32+g*4+b].b:=b*16;
         end;
palette256(Palette);
for i:=0 to 255
do FillPoly(6,Esa,i);
while readkey='' do;
CloseGraph;
writeln('Again in text mode...');
end.

}
end.