{***************************************************

 function description:

 revision information:
 date   rev   information
 12/97  0.1   created from 3D.PAS for DOS
 ***************************************************}
{define collision}
{ simboli riconosciuti dal compilatore:                         }
{       collision  abilita il controllo di collisione           }

program Win3D;

{$R WIN3D.RES}

uses
  WVGA256, File3d, Data3d, Graph3d, WinDos,
  WinTypes, WinProcs, WObjects, Strings, CommDlg;

const
  idm_Load    = 100;
  idm_About   = 103;
  idm_Info    = 104;

  idm_320     = 201;
  idm_640     = 202;

  acc_ctrl_S  = 901; (* ^S *)
  acc_ctrl_L  = 902; (* ^L *)
  acc_ctrl_R  = 903; (* ^R *)
  acc_ctrl_P  = 904; (* ^P *)

  ININAME     = 'WIN3D.INI';

type
  TFileName = array[0..fsPathName] of Char;

  TApp = object(TApplication)
    procedure InitMainWindow; virtual;
    procedure InitInstance; virtual;
  end;

  P3DWindow = ^T3DWindow;
  T3DWindow = object(TGraph3DWindow)
    scene: pnode;
    FileName: TFileName;
    xo,yo,zo:real; { posizione dell' osservatore }
    phi:real; { direzione di visione in radianti }
    linstep:real;
    log:(null, play, rec);
    logf:text;

    constructor Init(AParent: PWindowsObject; Title: PChar);
    destructor Done; virtual;
    procedure SetUpWindow; virtual;
    procedure SetSize320(var Message: TMessage); Virtual cm_first + idm_320;
    procedure SetSize640(var Message: TMessage); Virtual cm_first + idm_640;
    procedure Info(var Message: TMessage); Virtual cm_first + idm_Info;
    procedure About(var Message: TMessage); Virtual cm_first + idm_About;
    procedure UpdataOPosition(newxo,newyo,newzo:real;p:pnode);
    procedure ResetPosition;
    procedure GetWindowClass(var WndClass: TWndClass); virtual;
    procedure LoadFile(var Message: TMessage); virtual cm_first + idm_Load;
    procedure WMKeyDown(var Msg: TMessage); virtual wm_First + wm_KeyDown;
    procedure WMTimer(var Msg: TMessage); virtual WM_Timer;
    procedure SavePos(var Message: TMessage); virtual cm_first + acc_ctrl_S;
    procedure RestorePos(var Message: TMessage); virtual cm_first + acc_ctrl_L;
    procedure OpenLog(var Message: TMessage); virtual cm_first + acc_ctrl_R;
    procedure CloseLog;
    procedure PlayLog(var Message: TMessage); virtual cm_first + acc_ctrl_P;
  end;

  PInfoDialog = ^TInfoDialog;
  TInfoDialog = object(TDialog)
    procedure SetUpWindow; virtual;
  end;

{ T3DWindow }

constructor T3DWindow.Init(AParent: PWindowsObject; Title: PChar);
begin
  TGraph3DWindow.Init(AParent, Title);
  scene:=nil;
  log:=null;
  ResetPosition;
  StrCopy(FileName,'');
end;

procedure T3DWindow.ResetPosition;
begin
  phi:=0;
  xo:=1.2;
  yo:=0;
  zo:=1.1;
  linstep:=0;
end;

destructor T3DWindow.Done;
begin
  KillTimer(HWindow,1);
  if log <> null
  then CloseLog;
  TGraph3DWindow.Done;
end;

procedure T3DWindow.SetUpWindow;
begin
  SetTimer(HWindow,1,333,nil); (* one third of second *)
end;

procedure T3DWindow.SetSize320(var Message: TMessage);
begin
  CheckMenuItem(GetMenu(HWindow), idm_320, mf_Checked or mf_ByCommand);
  CheckMenuItem(GetMenu(HWindow), idm_640, mf_UnChecked or mf_ByCommand);
  SetWindowSize(z320);
end;

procedure T3DWindow.SetSize640(var Message: TMessage);
begin
  CheckMenuItem(GetMenu(HWindow), idm_640, mf_Checked or mf_ByCommand);
  CheckMenuItem(GetMenu(HWindow), idm_320, mf_UnChecked or mf_ByCommand);
  SetWindowSize(z640);
end;

procedure T3DWindow.LoadFile(var Message: TMessage);
const
  DefExt = 'dat';
var
  OpenFN      : TOpenFileName;
  FullFileName: TFileName;
  Dir: array[0..fsDirectory] of Char;
begin
  StrCopy(FullFileName, '');
  FillChar(OpenFN, SizeOf(TOpenFileName), #0);
  with OpenFN do
  begin
    hInstance     := HInstance;
    hwndOwner     := HWindow;
    lpstrDefExt   := DefExt;
    lpstrFile     := FullFileName;
    lpstrFilter   := '3D/dat files'+#0+'*.dat'+#0+#0;
    flags         := ofn_FileMustExist(* or ofn_AllowMultiSelect*);
    lStructSize   := sizeof(TOpenFileName);
    nFilterIndex  := 1;       {Index into Filter String in lpstrFilter}
    nMaxFile      := SizeOf(FullFileName);
  end;
  GetCurDir(Dir,0);
  GetPrivateProfileString('Files','LastDir',Dir,Dir,sizeof(Dir),IniName);
  SetCurDir(Dir);
  if GetOpenFileName(OpenFN)
  then begin
       readfile(scene,StrPas(FullFileName));
       StrCopy(FileName, FullFileName);
       ResetPosition;
       show(scene,xo,yo,zo,phi,1.57);
       EnableMenuItem(GetMenu(HWindow), idm_Load, mf_Grayed);
       GetCurDir(Dir,0);
       WritePrivateProfileString('Files','LastDir',Dir,IniName);
       end;
end;


(* the following procedure was used just for testing WVGA256

procedure T3DWindow.Example(var Message: TMessage);
const
 Esa: array[1..6] of tPoint = ((X: 50; Y: 0),(X:150; Y: 0),
   (X: 200; Y: 50), (X: 150; Y: 100), (X: 50; Y: 100), (X: 0; Y: 50));
var
  i,j: Integer;
  Palette:PaletteType;
begin
  for i:=0 to 15
  do for j:=0 to 15
     do begin
           Palette[i+j*16].r:=i*4;
           Palette[i+j*16].g:=i*4;
           Palette[i+j*16].b:=j*4;
        end;
  palette256(Palette);
  for i:=0 to 31
  do for j:=0 to 19
     do PutPixel(i * 10, j * 10, i);

  for i:=0 to 255
  do Line(i,0,255-i,199,i);

  for i:=0 to 200
  do begin
     for j:=1 to 6
     do begin
          inc(Esa[j].x);
          inc(Esa[j].y)
        end;
     SetFillColor(i);
     FillPoly(6,Esa);
     end;

  for i:=0 to 15
  do for j:=0 to 15
     do Bar(j*10,i*10,j*10+8,i*10+8,i+j*16);
  TransferImage;
end;
*)

procedure T3DWindow.About(var Message: TMessage);
begin
  Application^.ExecDialog(new(PDialog,init(@Self,'About')));
end;

procedure T3DWindow.GetWindowClass(var WndClass: TWndClass);
begin
  TWindow.GetWindowClass(WndClass);
  WndClass.hIcon := LoadIcon(HInstance, 'ICONA');
  WndClass.lpszMenuName := 'Menu';
end;

procedure T3DWindow.UpdataOPosition(newxo, newyo, newzo:real; p:pnode);
var new:tpoint;
begin
{$ifdef collision}
  new.x:=newxo;
  new.y:=newyo;
  new.z:=newzo;
  if PointIsIncluded(new, p, phi) = nil
  then begin
{$endif}
         xo:=newxo;
         yo:=newyo;
         zo:=newzo;
{$ifdef collision}
       end
  else linstep:=0;
{$endif}
end;

procedure T3DWindow.SavePos(var Message: TMessage);
var f:text;
begin
  assign(f,'sav3Dpos.log');
  rewrite(f);
  writeln(f,phi,' ',xo,' ',yo,' ',zo);
  close(f);
end;

procedure T3DWindow.RestorePos(var Message: TMessage);
var f:text;
begin
  assign(f,'sav3Dpos.log');
(*$I-*)
  reset(f);
  if IOResult <> 0
  then exit;
(*$I+*)
  readln(f,phi,xo,yo,zo);
  close(f);
  linstep:=0;
  show(scene,xo,yo,zo,phi,1.57);
end;

procedure T3DWindow.OpenLog(var Message: TMessage);
begin
  if log <> null
  then CloseLog;
  assign(logf,'log3Dpos.log');
  rewrite(logf);
  writeln(logf,phi,' ',xo,' ',yo,' ',zo);
  log:=rec;
end;

procedure T3DWindow.CloseLog;
begin
  close(logf);
  log:=null;
end;

procedure T3DWindow.PlayLog(var Message: TMessage);
begin
  if log <> null
  then CloseLog;
  assign(logf,'log3Dpos.log');
(*$I-*)
  reset(logf);
  if IOResult <> 0
  then exit;
(*$I+*)
  log:=play;
  if not eof(logf)
  then begin
       readln(logf, phi, xo, yo, zo);
       show(scene, xo, yo, zo, phi, 1.57);
       end;
end;

procedure T3DWindow.WMKeyDown(var Msg: TMessage);
const phistep=0.1;  { step della direzione dell'osservatore  }
var change: boolean;

  function Shifted:boolean;
  begin
    Shifted:=GetKeyState(VK_SHIFT) and $8000 <> 0;
  end;

begin
  change:=true;
  case Msg.WParam of
    VK_UP: if Shifted
           then UpdataOPosition(xo, yo, zo+0.2, scene)
           else linstep:=linstep + 0.03;
    VK_DOWN: if Shifted
           then UpdataOPosition(xo, yo, zo-0.2, scene)
           else linstep:=linstep - 0.03;
    VK_LEFT: phi:=phi + phistep;
    VK_RIGHT: phi:=phi - phistep;
  else change:=false;
  end; (* of case *)
  if change
  then begin
       UpdataOPosition(xo+linstep*cos(phi), yo+linstep*sin(phi), zo, scene);
       show(scene, xo, yo, zo, phi, 1.57);
       if log = rec
       then writeln(logf, phi, ' ', xo, ' ', yo, ' ', zo);
       end;
end;

procedure T3DWindow.WMTimer(var Msg: TMessage);
begin
  if log = play
  then if eof(logf)
       then CloseLog
       else begin
            readln(logf, phi, xo, yo, zo);
            show(scene, xo, yo, zo, phi, 1.57);
            end
  else begin
       linstep:=0.95 * linstep;
       if abs(linstep) < 0.01
       then linstep:=0
       else begin
            UpdataOPosition(xo+linstep*cos(phi), yo+linstep*sin(phi), zo, scene);
            show(scene, xo, yo, zo, phi, 1.57);
            if log = rec
            then writeln(logf, phi, ' ', xo, ' ', yo, ' ', zo);
            end;
       end;
end;

procedure T3DWindow.Info(var Message: TMessage);
begin
  Application^.ExecDialog(new(PInfoDialog,init(@Self,'Info')));
end;

procedure TInfoDialog.SetUpWindow;
begin
  TDialog.SetUpWindow;
  if P3DWindow(Parent)^.scene <> nil
  then SetDlgItemText(HWindow,2,P3DWindow(Parent)^.FileName);
end;

{ TApp }

procedure TApp.InitMainWindow;
begin
  MainWindow := New(P3DWindow, Init(nil, 'Win3D/01'));
end;

procedure TApp.InitInstance;
begin
  TApplication.InitInstance;
  HAccTable := LoadAccelerators(HInstance, 'ACCELERATORS');
end;

var
  App: TApp;
begin
  App.Init('Win3D');
  App.Run;
  App.Done;
end.
